import unittest

import numpy as np
import pandas as pd
import statsmodels.api as sm
from scipy import stats
from scipy.special import expit

from ananke.estimation import CausalEffect
from ananke.graphs import ADMG

TOL = 0.1

# print(stats.describe(Y_2))


class TestCounterfactualMean(unittest.TestCase):

    # general check: a-fixablility + mb-shielded ADMG
    def test_a_fixability(self):

        vertices = ["Z1", "Z2", "C1", "C2", "T", "M", "Y", "D1", "D2"]
        di_edges = [
            ("C1", "Z1"),
            ("C1", "T"),
            ("C1", "M"),
            ("C2", "Z1"),
            ("C2", "T"),
            ("C2", "M"),
            ("Z1", "Z2"),
            ("Z2", "T"),
            ("T", "M"),
            ("M", "Y"),
            ("M", "D1"),
            ("Y", "D2"),
            ("D1", "D2"),
        ]
        bi_edges = [("Z1", "T"), ("Z2", "C1"), ("C2", "Y"), ("D1", "Y")]
        G = ADMG(vertices, di_edges, bi_edges)

        data = pd.DataFrame()
        ace = CausalEffect(G, "T", "Y")

        self.assertFalse(ace.is_mb_shielded)
        self.assertEqual(ace.strategy, "a-fixable")
        with self.assertRaises(RuntimeError):
            ace.compute_effect(data, "eff-aipw")

    # a-fixablity: no C and binary/continuous Y
    def test_ignorable_model(self):
        np.random.seed(0)
        vertices = ["T", "Y"]
        di_edges = [("T", "Y")]
        bi_edges = []
        G = ADMG(vertices, di_edges, bi_edges)

        size = 5000
        T = np.random.binomial(1, 0.3, size)

        # First dataset: Continuous Y
        ace_truth_1 = 0.4
        Y_1 = -0.2 + ace_truth_1 * T + np.random.normal(0, 1, size)
        data_1 = pd.DataFrame({"T": T, "Y": Y_1})

        ace_1 = CausalEffect(G, "T", "Y")
        ace_1_ipw = ace_1.compute_effect(data_1, "ipw")
        ace_1_gformula = ace_1.compute_effect(data_1, "gformula")
        ace_1_aipw = ace_1.compute_effect(data_1, "aipw")
        ace_1_eff_aipw = ace_1.compute_effect(data_1, "eff-aipw")

        self.assertTrue(ace_1.is_mb_shielded)
        self.assertEqual(ace_1.strategy, "a-fixable")
        self.assertTrue(abs(ace_1_ipw - ace_truth_1) < TOL)
        self.assertTrue(abs(ace_1_gformula - ace_truth_1) < TOL)
        self.assertTrue(abs(ace_1_aipw - ace_truth_1) < TOL)
        self.assertTrue(abs(ace_1_eff_aipw - ace_truth_1) < TOL)

        # Second dataset: Binary Y
        ace_truth_2 = -0.4
        p_y = expit(0.8 + ace_truth_2 * T)
        Y_2 = np.random.binomial(1, p_y, size)
        data_2 = pd.DataFrame({"T": T, "Y": Y_2})

        ace_2 = CausalEffect(G, "T", "Y")
        ace_2_gformula = ace_2.compute_effect(data_2, "gformula")
        ace_2_aipw = ace_2.compute_effect(data_2, "aipw")

        self.assertTrue(abs(ace_2_gformula - ace_truth_2) < TOL)
        self.assertTrue(abs(ace_2_aipw - ace_truth_2) < TOL)

    # a-fixable: continuous Y
    def test_a_fixability_compute_ace(self):
        np.random.seed(0)

        vertices = ["Z1", "Z2", "C1", "C2", "T", "M", "Y"]
        di_edges = [
            ("C1", "Z1"),
            ("C1", "T"),
            ("C1", "M"),
            ("C2", "Z1"),
            ("C2", "T"),
            ("C2", "M"),
            ("Z1", "Z2"),
            ("Z2", "T"),
            ("T", "M"),
            ("M", "Y"),
        ]
        bi_edges = [("Z1", "T"), ("Z2", "C1"), ("C2", "Y")]
        G = ADMG(vertices, di_edges, bi_edges)

        size = 2000
        U1 = np.random.binomial(1, 0.4, size)
        U2 = np.random.uniform(0, 1.5, size)
        U3 = np.random.binomial(1, 0.6, size)
        U4 = np.random.uniform(-1, 0.2, size)
        U5 = np.random.binomial(1, 0.3, size)
        U6 = np.random.uniform(0.5, 1.5, size)

        C1 = U3 + U4 + np.random.normal(0, 1, size)
        C2 = U5 * U6 + np.random.normal(0, 1, size)

        p_z1 = expit(0.4 + 0.3 * C1 - 0.4 * C2 - 0.5 * U1 * U2)
        Z1 = np.random.binomial(1, p_z1, size)

        Z2 = 1 + Z1 + U3 + U4 + np.random.normal(0, 1, size)

        p_t1 = expit(0.5 - 0.3 * C1 - 0.4 * C2 + 0.3 * U1 - 0.3 * U2)
        T = np.random.binomial(1, p_t1, size)

        M = 1 + 0.5 * C1 - 0.8 * C2 - 0.5 * T + np.random.normal(0, 1, size)
        Y = 1 + 1 * M + U5 + U6 + np.random.normal(0, 1, size)

        data = pd.DataFrame(
            {"C1": C1, "C2": C2, "Z1": Z1, "Z2": Z2, "T": T, "M": M, "Y": Y}
        )

        ace_truth = -0.5

        ace = CausalEffect(G, "T", "Y")
        ace_ipw = ace.compute_effect(data, "ipw")
        ace_gformula = ace.compute_effect(data, "gformula")
        ace_aipw = ace.compute_effect(data, "aipw")
        ace_eff = ace.compute_effect(data, "eff-aipw")

        self.assertTrue(abs(ace_ipw - ace_truth) < TOL)
        self.assertTrue(abs(ace_gformula - ace_truth) < TOL)
        self.assertTrue(abs(ace_aipw - ace_truth) < TOL)
        self.assertTrue(abs(ace_eff - ace_truth) < TOL)

    # general check: p-fixablility + mb-shielded ADMG
    def test_p_fixability(self):

        # First graph: mb-shielded
        vertices_1 = ["C", "T", "M", "L", "Y"]
        di_edges_1 = [
            ("C", "T"),
            ("C", "M"),
            ("C", "L"),
            ("C", "Y"),
            ("T", "M"),
            ("M", "L"),
            ("M", "Y"),
            ("L", "Y"),
        ]
        bi_edges_1 = [("T", "L"), ("T", "Y")]
        G_1 = ADMG(vertices_1, di_edges_1, bi_edges_1)

        ace_1 = CausalEffect(G_1, "T", "Y")
        self.assertTrue(ace_1.is_mb_shielded)
        self.assertEqual(ace_1.strategy, "p-fixable")

        # Second graph: mb-shielded
        vertices_2 = ["C", "T", "M", "L", "Y"]
        di_edges_2 = [
            ("C", "T"),
            ("C", "M"),
            ("C", "L"),
            ("C", "Y"),
            ("T", "M"),
            ("M", "L"),
            ("T", "Y"),
            ("L", "Y"),
        ]
        bi_edges_2 = [("T", "L"), ("M", "Y")]
        G_2 = ADMG(vertices_2, di_edges_2, bi_edges_2)

        ace_2 = CausalEffect(G_2, "T", "Y")
        self.assertTrue(ace_2.is_mb_shielded)
        self.assertEqual(ace_2.strategy, "p-fixable")

        # Third graph: not mb-shielded
        vertices_3 = ["C1", "C2", "Z1", "Z2", "T", "M", "L", "Y"]
        di_edges_3 = [
            ("C1", "T"),
            ("C1", "L"),
            ("C2", "M"),
            ("C2", "L"),
            ("C2", "Y"),
            ("T", "M"),
            ("M", "L"),
            ("L", "Y"),
        ]
        bi_edges_3 = [("Z1", "C1"), ("Z2", "C2"), ("T", "L")]
        G_3 = ADMG(vertices_3, di_edges_3, bi_edges_3)

        data = pd.DataFrame()
        ace_3 = CausalEffect(G_3, "T", "Y")

        self.assertFalse(ace_3.is_mb_shielded)
        self.assertEqual(ace_3.strategy, "p-fixable")
        with self.assertRaises(RuntimeError):
            ace_3.compute_effect(data, "eff-apipw")

    # p-fixable: Binary M, Binary L, Continuous Y \not\in {L, M}
    def test_p_fixability_binaryML(self):
        np.random.seed(0)
        vertices = ["C1", "C2", "Z1", "Z2", "T", "M", "L", "Y"]
        di_edges = [
            ("C1", "T"),
            ("C1", "L"),
            ("C2", "T"),
            ("C2", "M"),
            ("C2", "L"),
            ("C2", "Y"),
            ("T", "M"),
            ("M", "L"),
            ("L", "Y"),
        ]
        bi_edges = [("Z1", "C1"), ("Z2", "C2"), ("T", "L")]
        G = ADMG(vertices, di_edges, bi_edges)

        size = 1000
        U1 = np.random.binomial(1, 0.4, size)
        U2 = np.random.uniform(0, 1.5, size)
        U3 = np.random.binomial(1, 0.6, size)
        U4 = np.random.uniform(-1, 0.2, size)
        U5 = np.random.binomial(1, 0.3, size)
        U6 = np.random.uniform(0.5, 1.5, size)

        p_z1 = expit(0.4 - U1 + U2)
        Z1 = np.random.binomial(1, p_z1, size)

        p_c1 = expit(-0.1 + U1 - U2)  # + 0.5*Z1)
        C1 = np.random.binomial(1, p_c1, size)

        C2 = 1 + U3 - U4 + np.random.normal(0, 1, size)

        Z2 = -0.5 + U3 - U4 + np.random.normal(0, 1, size)

        p_t = expit(0.5 + 0.5 * C1 - 0.4 * C2 - 0.4 * U5 + 0.4 * U6)
        T = np.random.binomial(1, p_t, size)

        p_m = expit(-0.3 + 1.5 * T - 0.3 * C2)
        M = np.random.binomial(1, p_m, size)

        p_l = expit(0.75 - 0.8 * M - 0.4 * C1 - 0.3 * C2 - 0.4 * U5 + 0.5 * U6)
        L = np.random.binomial(1, p_l, size)

        Y = 1 + 1 * L + C2 + np.random.normal(0, 1, size)

        data = pd.DataFrame(
            {
                "C1": C1,
                "C2": C2,
                "Z1": Z1,
                "Z2": Z2,
                "T": T,
                "M": M,
                "L": L,
                "Y": Y,
            }
        )

        # Compute the true ACE
        # vertices_hidden = ['C1', 'C2', 'Z1', 'Z2', 'T', 'M', 'L', 'Y', 'U1', 'U2', 'U3', 'U4', 'U5', 'U6']
        # di_edges_hidden = [('C1', 'T'), ('C1', 'L'), ('C2', 'T'), ('C2', 'M'),
        #                    ('C2', 'L'), ('C2', 'Y'), ('T', 'M'), ('M', 'L'), ('L', 'Y'),
        #                    ('U1', 'Z1'), ('U1', 'C1'), ('U2', 'Z1'), ('U2', 'C1'),
        #                    ('U3', 'C2'), ('U3', 'Z2'), ('U4', 'C2'), ('U4', 'Z2'),
        #                    ('U5', 'T'), ('U5', 'L'), ('U6', 'T'), ('U6', 'L')]
        # G_hidden = ADMG(vertices_hidden, di_edges_hidden, [])
        # data_hidden = pd.DataFrame({'C1': C1, 'C2': C2, 'Z1': Z1, 'Z2': Z2, 'T': T, 'M': M, 'L': L, 'Y': Y,
        #                             'U1': U1, 'U2': U2, 'U3': U3, 'U4': U4, 'U5': U5, 'U6': U6})
        # ace_hidden = AverageCausalEffect(G_hidden, 'T', 'Y')
        # ace_hidden_ipw, _, _ = ace_hidden.bootstrap_ace(data_hidden, "ipw")
        # ace_hidden_gformula, _, _ = ace_hidden.bootstrap_ace(data_hidden, "gformula")
        # ace_hidden_aipw, _, _ = ace_hidden.bootstrap_ace(data_hidden, "aipw")
        # ace_hidden_eff, _, _ = ace_hidden.bootstrap_ace(data_hidden, "eff-aipw")
        #
        # print(ace_hidden_ipw)
        # print(ace_hidden_gformula)
        # print(ace_hidden_aipw)
        # print(ace_hidden_eff, "\n")
        ace_truth = -0.07

        ace = CausalEffect(G, "T", "Y")
        ace_pipw = ace.compute_effect(data, "p-ipw")
        ace_dipw = ace.compute_effect(data, "d-ipw")
        ace_apipw = ace.compute_effect(data, "apipw")
        ace_eff = ace.compute_effect(data, "eff-apipw")

        self.assertTrue(abs(ace_pipw - ace_truth) < TOL)
        self.assertTrue(abs(ace_dipw - ace_truth) < TOL)
        self.assertTrue(abs(ace_apipw - ace_truth) < TOL)
        self.assertTrue(abs(ace_eff - ace_truth) < TOL)
        with self.assertRaises(RuntimeError):
            ace.compute_effect(data, "ipw")
        with self.assertRaises(RuntimeError):
            ace.compute_effect(data, "gformula")
        with self.assertRaises(RuntimeError):
            ace.compute_effect(data, "aipw")
        with self.assertRaises(RuntimeError):
            ace.compute_effect(data, "eff-aipw")

    # p-fixable: Continuous M, Continuous L, continuous Y \not\in {L, M}
    def test_p_fixability_continuousML(self):
        np.random.seed(0)
        vertices = ["C1", "C2", "Z1", "Z2", "T", "M", "L", "Y"]
        di_edges = [
            ("C1", "T"),
            ("C1", "L"),
            ("C2", "T"),
            ("C2", "M"),
            ("C2", "L"),
            ("C2", "Y"),
            ("T", "M"),
            ("M", "L"),
            ("L", "Y"),
        ]
        bi_edges = [("Z1", "C1"), ("Z2", "C2"), ("T", "L")]
        G = ADMG(vertices, di_edges, bi_edges)

        size = 5000
        U1 = np.random.binomial(1, 0.4, size)
        U2 = np.random.uniform(0, 1.5, size)
        U3 = np.random.binomial(1, 0.6, size)
        U4 = np.random.uniform(-1, 0.2, size)
        U5 = np.random.binomial(1, 0.3, size)
        U6 = np.random.uniform(0.5, 1.5, size)

        p_z1 = expit(0.4 - U1 + U2)
        Z1 = np.random.binomial(1, p_z1, size)

        p_c1 = expit(-0.1 + U1 - U2)
        C1 = np.random.binomial(1, p_c1, size)

        C2 = 1 + U3 - U4 + np.random.normal(0, 1, size)
        Z2 = -0.5 + U3 - U4 + np.random.normal(0, 1, size)

        p_t = expit(0.5 + 0.5 * C1 - 0.4 * C2 - 0.4 * U5 + 0.4 * U6)
        T = np.random.binomial(1, p_t, size)

        M = -0.3 + 0.8 * T - 0.3 * C2 + np.random.normal(0, 1, size)
        L = (
            0.75
            - 1.5 * M
            - 0.4 * C1
            - 0.3 * C2
            - 0.4 * U5
            + 0.5 * U6
            + np.random.normal(0, 1, size)
        )
        Y = 1 + 1 * L + C2 + np.random.normal(0, 1, size)

        data = pd.DataFrame(
            {
                "C1": C1,
                "C2": C2,
                "Z1": Z1,
                "Z2": Z2,
                "T": T,
                "M": M,
                "L": L,
                "Y": Y,
            }
        )

        # Compute the true ACE = (coefficient of T in M)*(coefficient of M in L)
        ace_truth = -1.2

        ace = CausalEffect(G, "T", "Y")
        ace_pipw = ace.compute_effect(data, "p-ipw")
        ace_dipw = ace.compute_effect(data, "d-ipw")

        print(ace_pipw)
        print(ace_dipw)

        self.assertTrue(abs(ace_pipw - ace_truth) < TOL)
        self.assertTrue(abs(ace_dipw - ace_truth) < TOL)

    # p-fixable: Continuous M, no C , continuous Y \in {L} AND continuous Y \in {M}
    def test_p_fixable_continuousY_inML(self):
        np.random.seed(0)
        vertices = ["T", "M", "Y"]
        di_edges = [("T", "M"), ("M", "Y")]
        bi_edges_1 = [("T", "Y")]
        bi_edges_2 = [("M", "Y")]

        # First graph: Y \in L (front-door), mb-shielded
        G_1 = ADMG(vertices, di_edges, bi_edges_1)

        # Second graph: Y \in M, not mb-shielded
        G_2 = ADMG(vertices, di_edges, bi_edges_2)

        size = 5000
        U1 = np.random.binomial(1, 0.4, size)
        U2 = np.random.uniform(0, 1.5, size)

        p_t = expit(0.8 + 0.4 * U1 - 0.8 * U2)
        T = np.random.binomial(1, p_t, size)

        M = -0.3 + 0.8 * T + np.random.normal(0, 1, size)
        Y = 1 + 1 * M + 0.5 * U1 + 0.3 * U2 + np.random.normal(0, 1, size)

        data = pd.DataFrame({"T": T, "M": M, "Y": Y})

        # Compute the true ACE = (coefficient of T in M)*(coefficient of M in L)
        ace_truth = 0.8

        # Test on the first graph
        ace_1 = CausalEffect(G_1, "T", "Y")
        ace_1_pipw, _, _ = ace_1.compute_effect(data, "p-ipw", n_bootstraps=1)
        ace_1_dipw = ace_1.compute_effect(data, "d-ipw")
        ace_1_apipw = ace_1.compute_effect(data, "apipw")
        ace_1_eff = ace_1.compute_effect(data, "eff-apipw")

        self.assertTrue(ace_1.is_mb_shielded)
        self.assertTrue(abs(ace_1_pipw - ace_truth) < TOL)
        self.assertTrue(abs(ace_1_dipw - ace_truth) < TOL)
        self.assertTrue(abs(ace_1_apipw - ace_truth) < TOL)
        self.assertTrue(abs(ace_1_eff - ace_truth) < TOL)

        # Test on the second graph
        ace_2 = CausalEffect(G_2, "T", "Y")
        ace_2_pipw = ace_2.compute_effect(data, "p-ipw")
        ace_2_dipw = ace_2.compute_effect(data, "d-ipw")
        ace_2_apipw = ace_2.compute_effect(data, "apipw")

        self.assertFalse(ace_2.is_mb_shielded)
        self.assertTrue(abs(ace_2_pipw - ace_truth) < TOL)
        self.assertTrue(abs(ace_2_dipw - ace_truth) < TOL)
        self.assertTrue(abs(ace_2_apipw - ace_truth) < TOL)

    # p-fixable: Continuous M, no C , binary Y \in {L} AND binary Y \in {M}
    def test_p_fixable_binaryY_inML(self):
        np.random.seed(0)
        vertices = ["T", "M", "Y"]
        di_edges = [("T", "M"), ("M", "Y")]
        bi_edges_1 = [("T", "Y")]
        bi_edges_2 = [("M", "Y")]

        # First graph: Y \in L (front-door)
        G_1 = ADMG(vertices, di_edges, bi_edges_1)

        # Second graph: Y \in M
        G_2 = ADMG(vertices, di_edges, bi_edges_2)

        size = 5000
        U1 = np.random.binomial(1, 0.4, size)
        U2 = np.random.uniform(0, 1.5, size)

        p_t = expit(0.5 - 0.4 * U1 + 0.4 * U2)
        T = np.random.binomial(1, p_t, size)

        M = -0.3 + 0.8 * T + np.random.normal(0, 1, size)

        p_y = expit(-1 + 0.4 * M + 0.5 * U1 + 0.3 * U2)
        Y = np.random.binomial(1, p_y, size)

        data = pd.DataFrame({"T": T, "M": M, "Y": Y})

        # Compute the true ACE (log of odds ratio)
        ace_truth = 0.32

        # Test the first graph
        ace_1 = CausalEffect(G_1, "T", "Y")
        ace_1_pipw = ace_1.compute_effect(data, "p-ipw")
        ace_1_dipw = ace_1.compute_effect(data, "d-ipw")

        self.assertTrue(abs(ace_1_pipw - ace_truth) < TOL)
        self.assertTrue(abs(ace_1_dipw - ace_truth) < TOL)

        # Test the second graph
        ace_2 = CausalEffect(G_2, "T", "Y")
        ace_2_pipw, _, _ = ace_2.compute_effect(data, "p-ipw", n_bootstraps=1)
        ace_2_dipw = ace_2.compute_effect(data, "d-ipw")

        self.assertTrue(abs(ace_2_pipw - ace_truth) < TOL)
        self.assertTrue(abs(ace_2_dipw - ace_truth) < TOL)

    # nested-fixable
    def test_nested_fixability(self):
        np.random.seed(0)
        vertices = ["C1", "C2", "T", "M", "Z", "R1", "R2", "Y"]
        di_edges = [
            ("C1", "T"),
            ("C1", "Y"),
            ("C2", "T"),
            ("C2", "Y"),
            ("R2", "Y"),
            ("Z", "T"),
            ("T", "R1"),
            ("T", "Y"),
            ("R1", "M"),
            ("M", "Y"),
        ]
        bi_edges = [
            ("Z", "R2"),
            ("T", "R2"),
            ("Z", "R1"),
            ("C1", "M"),
            ("C1", "Y"),
            ("C2", "M"),
            ("C2", "Y"),
        ]
        G = ADMG(vertices, di_edges, bi_edges)

        size = 2000
        U1 = np.random.binomial(1, 0.4, size)
        U2 = np.random.uniform(0, 0.5, size)
        U3 = np.random.binomial(1, 0.3, size)
        U4 = np.random.uniform(0, 0.5, size)
        U5 = np.random.binomial(1, 0.4, size)
        U6 = np.random.uniform(0, 1.5, size)
        U7 = np.random.binomial(1, 0.3, size)
        U8 = np.random.uniform(0, 0.5, size)
        U9 = np.random.binomial(1, 0.3, size)
        U10 = np.random.uniform(0, 0.5, size)

        p_r2 = expit(-0.2 + U1 - 0.8 * U2 + U3 + U4)
        R2 = np.random.binomial(1, p_r2, size)

        C1 = U7 - U8 + U9 + U10 + np.random.normal(0, 1, size)
        C2 = U7 - U8 + U9 * U10 + np.random.normal(0, 1, size)
        Z = U1 - U2 - U5 - U6 + np.random.normal(0, 1, size)

        p_t1 = expit(0.8 - 0.5 * C1 + 0.5 * C2 + 0.3 * Z + 0.5 * U3 - 0.4 * U4)
        T = np.random.binomial(1, p_t1, size)

        p_r1 = expit(0.2 + 0.7 * T - 0.6 * U5 - 0.6 * U6)
        R1 = np.random.binomial(1, p_r1, size)

        M = 1 - R1 - U7 + U8 + np.random.normal(0, 1, size)
        Y = (
            1
            + R2
            + M
            + 1 * T
            + C1
            + C2
            + U9
            + U10
            + np.random.normal(0, 1, size)
        )

        data = pd.DataFrame(
            {
                "C1": C1,
                "C2": C2,
                "R1": R1,
                "R2": R2,
                "Z": Z,
                "T": T,
                "M": M,
                "Y": Y,
            }
        )

        ace_truth = 0.8

        ace = CausalEffect(G, "T", "Y")
        ace_nipw = ace.compute_effect(data, "n-ipw")
        ace_anipw = ace.compute_effect(data, "anipw")

        self.assertEqual(ace.strategy, "nested-fixable")
        self.assertTrue(abs(ace_nipw - ace_truth) < TOL)
        self.assertTrue(abs(ace_anipw - ace_truth) < TOL)

        # ensure other methods don't work for this graph
        with self.assertRaises(RuntimeError):
            ace.compute_effect(data, "p-ipw")
        with self.assertRaises(RuntimeError):
            ace.compute_effect(data, "d-ipw")
        with self.assertRaises(RuntimeError):
            ace.compute_effect(data, "apipw")
        with self.assertRaises(RuntimeError):
            ace.compute_effect(data, "eff-apipw")

    def test_nested_fixable_Y_in_DT(self):
        np.random.seed(0)
        vertices = ["C", "Z1", "Z2", "T", "Y"]
        di_edges = [
            ("C", "T"),
            ("C", "Y"),
            ("Z1", "Z2"),
            ("Z2", "T"),
            ("T", "Y"),
        ]
        bi_edges = [("Z1", "T"), ("Z1", "Y")]
        G = ADMG(vertices, di_edges, bi_edges)

        size = 2000
        U1 = np.random.binomial(1, 0.4, size)
        U2 = np.random.uniform(0, 1.5, size)
        U3 = np.random.binomial(1, 0.3, size)
        U4 = np.random.uniform(0, 0.5, size)

        C = np.random.normal(0, 1, size)
        Z1 = U1 - U2 + 0.5 * U3 + U4 + np.random.normal(1, 1, size)

        p_z2 = expit(0.4 - 0.4 * Z1)
        Z2 = np.random.binomial(1, p_z2, size)

        p_t = expit(0.2 + 0.2 * C + 0.5 * Z2 + U1 - U2)
        T = np.random.binomial(1, p_t, size)

        Y = 1 + T - 0.6 * C + U3 + U4 + np.random.normal(0, 1, size)

        ace_truth = 1

        ace = CausalEffect(G, "T", "Y")
        data = pd.DataFrame({"Z1": Z1, "Z2": Z2, "C": C, "T": T, "Y": Y})
        ace_nipw = ace.compute_effect(data, "n-ipw")
        ace_anipw = ace.compute_effect(data, "anipw")

        self.assertTrue(abs(ace_nipw - ace_truth) < TOL)
        self.assertTrue(abs(ace_anipw - ace_truth) < TOL)

    # nested-fixablity: continuous variables
    def test_nested_fixable_Y_in_DT_continuousZ(self):
        np.random.seed(0)
        vertices = ["C", "Z1", "Z2", "T", "Y"]
        di_edges = [
            ("C", "T"),
            ("C", "Y"),
            ("Z1", "Z2"),
            ("Z2", "T"),
            ("T", "Y"),
        ]
        bi_edges = [("Z1", "T"), ("Z1", "Y")]
        G = ADMG(vertices, di_edges, bi_edges)

        size = 2000
        U1 = np.random.binomial(1, 0.4, size)
        U2 = np.random.uniform(0, 1.5, size)
        U3 = np.random.binomial(1, 0.3, size)
        U4 = np.random.uniform(0, 0.5, size)

        C = np.random.normal(0, 1, size)
        Z1 = U1 - U2 + 0.5 * U3 + U4 + np.random.normal(1, 1, size)
        Z2 = 0.4 - 0.4 * Z1 + np.random.normal(0, 1, size)

        p_t = expit(0.2 + 0.2 * C + 0.5 * Z2 + U1 - U2)
        T = np.random.binomial(1, p_t, size)

        Y = 1 + T - 0.6 * C + U3 + U4 + np.random.normal(0, 1, size)

        data = pd.DataFrame({"Z1": Z1, "Z2": Z2, "C": C, "T": T, "Y": Y})

        ace_truth = 1

        ace = CausalEffect(G, "T", "Y")
        ace.n_order = ["C", "Z1", "Z2", "T", "Y"]
        ace_nipw = ace.compute_effect(data, "n-ipw")
        ace_anipw = ace.compute_effect(data, "anipw")

        self.assertTrue(abs(ace_nipw - ace_truth) < TOL)
        self.assertTrue(abs(ace_anipw - ace_truth) < TOL)

    # nested-fixablity
    def test_nested_fixable_conditional_ignorability(self):
        np.random.seed(0)
        vertices = ["C", "T", "Y"]
        di_edges = [("C", "Y"), ("T", "Y")]
        bi_edges = [("C", "T")]
        G = ADMG(vertices, di_edges, bi_edges)

        size = 2000
        U1 = np.random.binomial(1, 0.4, size)
        U2 = np.random.uniform(0, 1.5, size)

        C = -1 + 0.4 * U1 + 0.8 * U2 + np.random.normal(0, 1, size)

        p_t = expit(0.2 + 0.2 * U1 + 0.5 * U2)
        T = np.random.binomial(1, p_t, size)

        p_y = expit(1 + 1.5 * T - 0.6 * C)
        Y = np.random.binomial(1, p_y, size)

        data = pd.DataFrame({"C": C, "T": T, "Y": Y})

        ace_truth = 1.5

        ace = CausalEffect(G, "T", "Y")
        ace_nipw = ace.compute_effect(data, "n-ipw")
        ace_anipw = ace.compute_effect(data, "anipw")

        self.assertTrue(abs(ace_nipw - ace_truth) < TOL)
        self.assertTrue(abs(ace_anipw - ace_truth) < TOL)

    # not ID
    def test_bow_arc(self):
        vertices = ["C", "T", "Y"]
        di_edges = [("C", "T"), ("C", "Y"), ("T", "Y")]
        bi_edges = [("T", "Y")]
        G = ADMG(vertices, di_edges, bi_edges)

        data = pd.DataFrame()

        ace = CausalEffect(G, "T", "Y")
        self.assertEqual(ace.strategy, "Not ID")
        with self.assertRaises(RuntimeError):
            ace.compute_effect(data, "n-ipw")
        with self.assertRaises(RuntimeError):
            ace.compute_effect(data, "anipw")


if __name__ == "__main__":
    unittest.main()



"""
Class for Intrinsic Graphs (IGs) -- a mixed graph used to compute intrinsic sets
of an ADMG in polynomial time.
"""

import itertools
import logging

from .graph import Graph

logger = logging.getLogger(__name__)


class IG(Graph):
    def __init__(self, admg):
        """
        Constructor.

        :param admg: ADMG object to calculate intrinsic sets for.
        """

        self.admg = admg
        self.iset_cadmg_map = (
            {}
        )  # dictionary mapping intrinsic sets to the reachable CADMG
        self.iset_fixing_order_map = {}
        super().__init__()

        # the IG is initialized with vertices corresponding to
        # reachable closures of singletons (these are guaranteed to be intrinsic)

        for v in admg.vertices:
            if not admg.vertices[v].fixed:
                rc, fixing_order, G = self.admg.reachable_closure([v])
                rc = frozenset(rc)
                self.add_vertex(rc)
                self.iset_cadmg_map[rc] = G
                self.iset_fixing_order_map[rc] = fixing_order

                # add di/bi edges that fulfill subset relation
                for i in self.vertices:

                    if i < rc:
                        self.add_diedge(i, rc)
                    elif rc < i:
                        self.add_diedge(rc, i)

                    if not (
                        i in self.ancestors([rc]) or rc in self.ancestors([i])
                    ) and self.bidirected_connected(i, rc):
                        self.add_biedge(i, rc)

    def bidirected_connected(self, s1, s2):
        """
        Check if two sets are bidirected connected in the original ADMG

        :param s1: First set corresponding to a vertex in the IG.
        :param s2: Second set corresponding to a vertex in the IG.
        :return: boolean corresponding to connectedness.
        """

        possible_endpoints = itertools.product(*[s1, s2])
        for combination in possible_endpoints:
            if self.admg.has_biedge(*combination):
                return True
        return False

    def maintain_subset_relation(self, s):
        """
        Add di edges to a newly inserted vertex s so as to maintain
        the subset relation.
        :param s: Frozenset corresponding to the new vertex.
        :return: None.
        """

        for i in self.vertices:
            if i < s:
                self.add_diedge(i, s)
            elif s < i:
                self.add_diedge(s, i)

    def add_new_biedges(self, s):
        """
        Naive O(``|I(G)|`` choose 2) implementation. Must ensure that biedges not added to ancestors.

        :param s: Frozen set corresponding to the new vertex.
        :return: None.
        """

        ancestors_s = self.ancestors([s])

        for i in self.vertices:

            if i not in ancestors_s and self.bidirected_connected(i, s):
                self.add_biedge(i, s)

    def merge(self, s1, s2):
        """
        Merge operation on two sets s1 and s2 to give a (possibly)
        new intrinsic set s3.

        :param s1: First set corresponding to a vertex in the IG.
        :param s2: Second set corresponding to a vertex in the IG.
        :return: None.
        """

        s3c = set(s1)
        s3c.update(set(s2))
        s3, fixing_order, G = self.admg.reachable_closure(s3c)
        s3 = frozenset(s3)
        self.delete_biedge(s1, s2)

        # if the intrinsic set already exists, ignore it
        if s3 in self.vertices:
            return

        # add the vertex and add di edges
        self.add_vertex(s3)
        self.iset_cadmg_map[s3] = G
        self.iset_fixing_order_map[s3] = fixing_order
        self.maintain_subset_relation(s3)

        # add new bi edges to the added vertex
        self.add_new_biedges(s3)

    def get_intrinsic_sets(self):
        """
        Get all intrinsic sets for given ADMG.


        :return:
        """

        # keep merging vertices that are connected
        # by bidirected edges to obtain new intrinsic
        # sets until there are none left
        while len(self.bi_edges) > 0:

            u, v = next(iter(self.bi_edges))
            self.merge(u, v)

        return set(self.vertices)

    def get_heads_tails(self):
        """
        Iterate through all intrinsic sets and get the recursive heads and tails.

        :return: List of tuples corresponding to (head, tail).
        """

        # make sure that the graph is fully initialized
        try:
            assert len(self.bi_edges) == 0

        except AssertionError:
            print(
                "There are still bidirected edges indicating more intrinsic sets can be merged"
            )

        heads_and_tails = []
        # go through each intrinsic set
        for iset in self.vertices:

            # look at the reachable CADMG to determine head/tail
            reachable_cadmg = self.iset_cadmg_map[iset]
            head = set()

            # heads are vertices that are childless in the corresponding reachable CADMG
            for v in iset:
                if len(reachable_cadmg.children([v])) == 0:
                    head.add(v)

            # tail is the remaining intrinsic set + parents of the intrinsic set
            tail = set(iset - head)
            tail = tail.union(reachable_cadmg.parents(iset))
            heads_and_tails.append((head, tail))

        return heads_and_tails



import unittest

from ananke.graphs import MissingADMG


class TestMissingADMG(unittest.TestCase):
    def test_draw_missing_admg(self):
        vertices = ["X_1", "Xp_1", "R_1", "X_2", "Xp_2", "R_2"]
        di_edges = [("R_1", "R_2"), ("X_1", "X_2")]

        G = MissingADMG(vertices=vertices, di_edges=di_edges)

        # check that no errors come up
        draw = G.draw()



import itertools

import numpy as np
import sympy as sp
from pgmpy.factors.discrete import TabularCPD

from ananke.factors import SymCPD, SymDiscreteFactor


class TestSymDiscreteFactor:
    def test_initialize(self):
        variables = ["A", "B"]
        cardinality = [2, 3]
        a = sp.Symbol("a")
        b = sp.Symbol("b")
        values = sp.Array([2, b + 1, b + 0.5, a - 3, b / a, b + a])
        result = SymDiscreteFactor(variables, cardinality, values)
        assert len(values) == 6
        assert result.values.shape == (2, 3)

    def test_marginalize(self):
        variables = ["A", "B"]
        cardinality = [2, 3]
        a = sp.Symbol("a")
        b = sp.Symbol("b")
        values = sp.Array([2, b + 1, b + 0.5, a - 3, b / a, b + a])
        result = SymDiscreteFactor(variables, cardinality, values)

        result = result.marginalize(["A"])
        truth = SymDiscreteFactor(
            ["B"], [3], sp.Array([2 + a - 3, b + 1 + b / a, b + 0.5 + b + a])
        )
        assert truth == result

    def test_product(self):
        a = sp.Symbol("a")
        b = sp.Symbol("b")
        c = sp.Symbol("c")
        d = sp.Symbol("d")

        variables1 = ["C", "A", "B"]
        cardinality1 = [2, 2, 3]
        # (0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2)
        values1 = sp.Array(
            [
                2,
                b + 1,
                b + 0.5,
                a - 3,
                b / a,
                b + a,
                2,
                b + 1,
                b + 0.5,
                a - 3,
                b / a,
                b + a,
            ]
        )
        factor1 = SymDiscreteFactor(variables1, cardinality1, values1)

        variables2 = ["D", "B", "C"]
        cardinality2 = [2, 3, 2]
        # (0, 0), (0, 1), (1, 0), (1, 1), (2, 0), (2, 1)
        values2 = sp.Array(
            [
                1,
                d,
                d**2,
                d**3,
                d**4,
                d**5,
                1,
                c,
                c**2,
                c**3,
                c**4,
                c**5,
            ]
        )
        factor2 = SymDiscreteFactor(variables2, cardinality2, values2)

        result = factor1.product(factor2)

        # 000, 001, 010, 011, 020, 021, 100, 101, 110, 111, 120, 121
        rvalues1 = values1.reshape(2, 2, 3)
        rvalues2 = values2.reshape(2, 3, 2)
        final_values = list()
        for i in itertools.product(*[range(2), range(3), range(2), range(2)]):
            final_values.append(
                rvalues1[i[2], i[0], i[1]] * rvalues2[i[3], i[1], i[2]]
            )
        truth = SymDiscreteFactor(
            variables=["A", "B", "C", "D"],
            cardinality=[2, 3, 2, 2],
            values=sp.Array(final_values),
        )
        assert result == truth

    def test_product_no_intersection(self):
        a = sp.Symbol("a")
        b = sp.Symbol("b")
        factor1 = SymDiscreteFactor(
            variables=["A"], cardinality=[2], values=sp.Array([a, a + 1])
        )
        factor2 = SymDiscreteFactor(
            variables=["B"], cardinality=[2], values=sp.Array([b, b + 1])
        )
        truth = SymDiscreteFactor(
            variables=["A", "B"],
            cardinality=[2, 2],
            values=sp.Array(
                [a * b, a * (b + 1), (a + 1) * b, (a + 1) * (b + 1)]
            ),
        )
        assert factor1.product(factor2) == truth

    def test_reduce(self):
        a = sp.Symbol("a")
        b = sp.Symbol("b")
        factor = SymDiscreteFactor(
            variables=["A", "B"],
            cardinality=[2, 2],
            values=sp.Array(
                [a * b, a * (b + 1), (a + 1) * b, (a + 1) * (b + 1)]
            ),
        )

        truth = SymDiscreteFactor(
            variables=["A"],
            cardinality=[2],
            values=sp.Array(
                [
                    a * b,
                    (a + 1) * b,
                ]
            ),
        )

        assert factor.reduce([("B", 0)]) == truth

    def test_divide(self):
        a = sp.Symbol("a")
        b = sp.Symbol("b")
        factor = SymDiscreteFactor(
            variables=["A", "B"],
            cardinality=[2, 2],
            values=sp.Array(
                [a * b, a * (b + 1), (a + 1) * b, (a + 1) * (b + 1)]
            ),
        )
        truth = SymDiscreteFactor(
            variables=["A", "B"],
            cardinality=[2, 2],
            values=sp.Array([1] * 4),
        )

        assert factor.divide(factor) == truth

    def test_marginalize_from_cpd(self):
        x1 = sp.Symbol("x1")
        y1 = sp.Symbol("y1")
        cpd_a = SymCPD(
            "A", 2, sp.Array([[x1, y1], [1 - x1, 1 - y1]]), ["B"], [2]
        )
        factor = cpd_a.to_factor()
        result = factor.marginalize(["B"])
        truth = SymDiscreteFactor(
            variables=["A"],
            cardinality=[2],
            values=sp.Array([x1 + y1, 2 - x1 - y1]),
        )

        result2 = factor.marginalize([])
        assert result2 == factor

    def test_empty_factor(self):
        a = sp.Symbol("a")
        factor = SymDiscreteFactor(
            variables=["A"], cardinality=[2], values=sp.Array([a, 1 - a])
        )
        result = factor.marginalize(["A"])

        truth = SymDiscreteFactor(variables=[], cardinality=[], values=1)

        assert result == truth
        assert factor.product(truth) == factor

    def test_get_value(self):
        x1 = sp.Symbol("x1")
        y1 = sp.Symbol("y1")
        cpd_a = SymCPD(
            "A", 2, sp.Array([[x1, y1], [1 - x1, 1 - y1]]), ["B"], [2]
        )

        result = cpd_a.get_value(A=0, B=1)
        truth = y1
        assert result == truth

    def test_to_pgmpy(self):
        cpd_a = SymCPD("A", 2, sp.Array([[1, 1], [0, 0]]), ["B"], [2])
        result = cpd_a.to_pgmpy()
        truth = TabularCPD("A", 2, np.array([[1, 1], [0, 0]]), ["B"], [2])

        assert result == truth



"""
Class for Linear Gaussian SEMs parametrized
by a matrix B representing regression coefficients
and a matrix Omega representing correlated errors
"""

import functools

import numpy as np
import pandas as pd
import statsmodels.api as sm
from scipy.optimize import minimize


def is_positive_definite(X):
    try:
        np.linalg.cholesky(X)
        return True
    except LinAlgError:
        return False


class LinearGaussianSEM:
    def __init__(self, graph):
        """
        Fits a linear Gaussian structural equation model to an :class:`ADMG` using the recursive iterative conditional fitting algorithm as proposed by Drton and Richardson.

        :param graph: graph (:class:`ADMG`) corresponding to the linear Gaussian SEM.
        """

        self.graph = graph
        # for a linear Gaussian SEM each edge is a parameter + noise for each vertex
        self.n_params = (
            len(graph.di_edges) + len(graph.bi_edges) + len(graph.vertices)
        )
        self._vertex_index_map = {v: i for i, v in enumerate(graph.vertices)}
        self._parent_index_map = {
            v: [self._vertex_index_map[p] for p in graph.parents([v])]
            for v in graph.vertices
        }
        self._sibling_index_map = {
            v: [self._vertex_index_map[s] for s in graph.siblings([v])]
            for v in graph.vertices
        }

        self.X_ = None  # data matrix
        self.S_ = None  # sample covariance matrix
        self.B_ = None  # direct edge coefficients
        self.omega_ = None  # correlation of errors

    def neg_loglikelihood(self, X):
        """
        Calculate log-likelihood of the data given the model.

        :param X: a N x M dimensional data matrix.
        :param weights: optional 1d numpy array with weights for each data point
                        (rows with higher weights are given greater importance).
        :return: a float corresponding to the log-likelihood.
        """
        assert self.B_ is not None

        S = np.cov(X.T)
        assert is_positive_definite(
            S
        ), "Sample covariance matrix is not positive definite"
        n, d = X.shape
        sigma = (
            np.linalg.inv(np.eye(d) - self.B_)
            @ (self.omega_)
            @ np.linalg.inv((np.eye(d) - self.B_).T)
        )
        # assert is_positive_definite(sigma)
        neg_lld = (n / 2) * (
            np.log(np.linalg.det(sigma))
            + np.trace(np.dot(np.linalg.inv(sigma), S))
        )

        return neg_lld

    def bic(self, X):
        """
        Calculate Bayesian information criterion of the data given the model.

        :param X: a N x M dimensional data matrix.
        :param weights: optional 1d numpy array with weights for each data point
                        (rows with higher weights are given greater importance).
        :return: a float corresponding to the log-likelihood.
        """

        return 2 * self.neg_loglikelihood(X) + np.log(len(X)) * self.n_params

    def fit(self, X, tol=1e-6, disp=None, standardize=False, max_iters=100):
        """
        Fit the model to data via (weighted) maximum likelihood estimation

        :param X: data -- a N x M dimensional pandas data frame.
        :param weights: optional 1d numpy array with weights for each data point
                        (rows with higher weights are given greater importance).
        :return: self.
        """

        # make copy of the data and standardize if necessary
        self.X_ = X.values.copy()
        n, d = self.X_.shape
        if standardize:
            self.X_ = (self.X_ - np.mean(self.X_, axis=0)) / np.std(
                self.X_, axis=0
            )

        # compute sample covariance
        self.S_ = np.cov(self.X_.T)

        # initialize B and omega
        self.B_, self.omega_ = np.zeros((d, d)), np.eye(d)

        # keep going until desired convergence
        cur_lld = self.neg_loglikelihood(self.X_)
        lld_diff = tol + 1
        n_iters = 0
        while np.abs(lld_diff) > tol or n_iters == max_iters:
            n_iters += 1

            for var in self.graph.vertices:

                # get index of the vertex
                var_index = self._vertex_index_map[var]

                # get epsilon_minusi
                epsilon = self.X_ - self.X_ @ self.B_.T
                epsilon_minusi = np.delete(epsilon, var_index, axis=1)

                # get omega and calculate pseudo variables Zi
                omega_minusi = np.delete(self.omega_, var_index, axis=0)
                omega_minusii = np.delete(omega_minusi, var_index, axis=1)
                omega_minusii_inv = np.linalg.inv(omega_minusii)

                # calculate Z_minusi
                Z_minusi = (omega_minusii_inv @ epsilon_minusi.T).T

                # insert a column of zeros to maintain the shape
                Z = np.insert(Z_minusi, var_index, 0, axis=1)

                Y = self.X_[:, var_index]
                parent_mat = self.X_[:, self._parent_index_map[var]]
                pseudo_mat = Z[:, self._sibling_index_map[var]]
                Xmat = np.hstack((np.ones((n, 1)), parent_mat, pseudo_mat))
                results = sm.OLS(Y, Xmat).fit()

                # need to now extract results and put it in the right spots
                i = 1
                for idx in self._parent_index_map[var]:
                    self.B_[var_index, idx] = results.params[i]
                    i += 1
                for idx in self._sibling_index_map[var]:
                    self.omega_[var_index, idx] = results.params[i]
                    self.omega_[idx, var_index] = results.params[i]
                    i += 1

                parent_mat = np.hstack((np.ones((n, 1)), parent_mat))
                _, cols = parent_mat.shape

                omega_i_minusi = np.delete(self.omega_, var_index, axis=1)[
                    var_index, :
                ]
                omega_minusi_i = np.delete(self.omega_, var_index, axis=0)[
                    :, var_index
                ]
                omega_ii = results.scale + (
                    omega_i_minusi @ omega_minusii_inv @ omega_minusi_i
                )
                self.omega_[var_index, var_index] = omega_ii
            new_lld = self.neg_loglikelihood(self.X_)
            lld_diff = new_lld - cur_lld

            cur_lld = new_lld

        return self

    def total_effect(self, A, Y):
        """
        Calculate the total causal effect of a set of treatments A on
        a set of outcomes Y.

        :param A: iterable corresponding to variable names that act as treatments.
        :param Y: iterable corresponding to variable names that act as outcomes.
        :return: a float corresponding to the total causal effect.
        """

        directed_paths = self.graph.directed_paths(A, Y)

        # if there are no directed paths, the effect is 0
        if len(directed_paths) == 0:
            return 0

        # otherwise do path analysis
        total_effect = 0
        for path in directed_paths:

            path_effect = 1
            for u, v in path:

                path_effect *= self.B_[
                    self._vertex_index_map[v], self._vertex_index_map[u]
                ]

            total_effect += path_effect

        return total_effect

    def draw(self, direction=None):
        """
        Visualize the graph.

        :return : dot language representation of the graph.
        """

        from graphviz import Digraph

        if self.B_ is None:
            raise AssertionError("Model must be fit before model can be drawn.")
        dot = Digraph()

        # set direction from left to right if that's preferred
        if direction == "LR":
            dot.graph_attr["rankdir"] = direction

        for v in self.graph.vertices.values():
            dot.node(v.name, shape="plaintext", height=".5", width=".5")

        for parent, child in self.graph.di_edges:
            i, j = self._vertex_index_map[child], self._vertex_index_map[parent]
            dot.edge(
                parent,
                child,
                color="blue",
                label=str(round(self.B_[i, j], 2)),
                fontsize="12",
            )
        for sib1, sib2 in self.graph.bi_edges:
            i, j = self._vertex_index_map[sib1], self._vertex_index_map[sib2]
            dot.edge(
                sib1,
                sib2,
                dir="both",
                color="red",
                label=str(round(self.omega_[i, j], 2)),
                fontsize="12",
            )

        return dot



"""
The code relies on the following papers:

[ER12] Evans, R. J., & Richardson, T. S. (2012). Maximum likelihood fitting of acyclic directed mixed graphs to binary data. arXiv preprint arXiv:1203.3479.
[ER19] Evans, R. J., & Richardson, T. S. (2019). Smooth, identifiable supermodels of discrete DAG models with latent variables. Bernoulli, 25(2), 848-876.

"""
import unittest
from collections import OrderedDict

import numpy as np
import pandas as pd
import pytest
from scipy.special import expit

from ananke.datasets import load_wisconsin_health_study
from ananke.graphs import ADMG, DAG
from ananke.models import binary_nested


def dag_chain_small():
    vertices = ["X_1", "X_2"]
    di_edges = [("X_1", "X_2")]
    G = DAG(di_edges=di_edges, vertices=vertices)

    return G


def wisconsin_fig8b_graph():
    G = ADMG(
        vertices=["X", "E", "M", "Y"],
        di_edges=[("X", "E"), ("E", "M"), ("X", "Y")],
        bi_edges=[("E", "Y")],
    )

    return G


def sixteen_not_eighteen_graph():
    vertices = ["A", "B", "C", "D", "E"]
    di_edges = [("A", "B"), ("B", "C"), ("C", "D")]
    bi_edges = [("E", "D"), ("B", "E")]

    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def standard_graph():
    """
    Fig 1 from [ER12]
    :return:
    """
    vertices = ["X1", "X2", "X3", "X4"]
    di_edges = [("X1", "X2"), ("X2", "X4")]
    bi_edges = [("X3", "X2"), ("X3", "X4")]
    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def standard_dag():
    vertices = ["X1", "X2", "X3", "X4"]
    di_edges = [("X1", "X2"), ("X2", "X3"), ("X3", "X4")]
    bi_edges = list()
    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def fig1b_graph():
    """
     Fig 1b from [ER19]

    :return:
    """
    vertices = ["X", "E", "M", "Y"]
    di_edges = [("X", "E"), ("E", "M"), ("M", "Y")]
    bi_edges = [("E", "Y")]
    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def fig2_graph():
    """
    Fig 2 from [ER19]

    :return:
    """
    vertices = ["X1", "X2", "X3", "X4"]
    di_edges = [("X1", "X3"), ("X2", "X4")]
    bi_edges = [("X1", "X4"), ("X2", "X3")]
    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def fig3_graph():
    """
    Fig 3 from [ER12]

    :return:
    """
    vertices = ["X1", "X2", "X3"]
    di_edges = [("X1", "X2")]
    bi_edges = [("X3", "X2"), ("X1", "X3")]
    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def fig7_graph():
    """
    Fig 7 from [ER19]
    :return:
    """
    vertices = ["2", "3", "4", "5"]
    di_edges = [("2", "3"), ("3", "4")]
    bi_edges = [("2", "5"), ("5", "4")]
    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def fig8_graph(k=5):
    """
    Fig 8 from [ER12]
    :param k:
    :return:
    """
    vertices = ["X{}".format(i) for i in range(k)] + [
        "Y{}".format(i) for i in range(k)
    ]
    di_edges = [("X{}".format(i), "Y{}".format(i)) for i in range(k)]
    bi_edges = [
        ("Y{}".format(i), "Y{}".format(i + 1)) for i in range(k - 1)
    ] + [("X{}".format(i), "X{}".format(i + 1)) for i in range(k - 1)]
    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def verma_graph(k=5):
    """
    Non-ancestral graph containing Verma constraints
    :param k:
    :return:
    """
    vertices = (
        ["A", "Y"]
        + ["L{}".format(i) for i in range(k)]
        + ["B{}".format(i) for i in range(k)]
    )
    di_edges = (
        [("A", "L0")]
        + [("L{}".format(i), "B{}".format(i)) for i in range(k)]
        + [("B{}".format(i), "Y") for i in range(k)]
    )
    bi_edges = [("L{}".format(i), "Y") for i in range(k)] + [
        ("B{}".format(i), "L{}".format(i + 1)) for i in range(k - 1)
    ]

    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def conditional_ignorability_graph():

    vertices = ["A", "C", "Y"]
    di_edges = [("C", "A"), ("C", "Y"), ("A", "Y")]
    G = DAG(di_edges=di_edges, vertices=vertices)

    return G


class TestFittingModels(unittest.TestCase):
    """
    This suite of tests checks maximum likelihood estimation works correctly in a variety of graphs
    """

    def test_estimate_likelihood(self):
        """
        Test that MLE works for a single district ancestral graph

        :return:
        """
        G = fig3_graph()
        districts = G.districts
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        q_vector = binary_nested.initialize_q_vector(intrinsic_dict)
        q = np.array(list(q_vector.values()))

        counts = np.array(
            [10, 3, 42, 4, 34, 9, 23, 124]
        )  # length of this vector is 2**len(vertices) = 8
        q_indices = binary_nested.compute_q_indices_by_district(
            q_vector_keys=list(q_vector.keys()), districts=districts
        )
        Ms, Ps = binary_nested.compute_all_M_and_P(
            G=G, intrinsic_dict=intrinsic_dict
        )
        prob = binary_nested._compute_prob(q, q_indices, districts, Ms, Ps)

        # Check probs sum to 1 and are not negative
        self.assertAlmostEqual(np.sum(prob), 1)
        self.assertTrue((prob >= 0).all())
        result = binary_nested._compute_likelihood(
            q, counts, q_indices, districts, Ms, Ps
        )

        # Check that likelihood is a number, and check that the probabilities are well formed
        self.assertTrue(~np.isnan(result))
        bnm = binary_nested.BinaryNestedModel(graph=G)
        bnm = bnm.fit(X=counts)
        final_q_vector = bnm.fitted_params
        final_p = binary_nested._compute_prob(
            np.array(list(final_q_vector.values())),
            q_indices,
            districts,
            Ms,
            Ps,
        )
        self.assertAlmostEqual(np.sum(final_p), 1)

    def test_that_dag_can_be_fitted(self):
        """
        Checks that the code works for DAGs, and consequently also checks that code works across more than one district.
        In the case of a DAG, we check that the empirical probabilities p(V = v) are recovered by the Mobius transform.

        :return:
        """
        G = standard_dag()
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        rng = np.random.default_rng(4213)
        counts = rng.integers(low=50, high=80, size=(2 ** len(G.vertices),))
        expected_prob = counts / counts.sum()

        bnm = binary_nested.BinaryNestedModel(graph=G)
        bnm = bnm.fit(X=counts)
        result = bnm.fitted_params
        q_indices = binary_nested.compute_q_indices_by_district(
            q_vector_keys=list(result.keys()), districts=G.districts
        )
        Ms, Ps = binary_nested.compute_all_M_and_P(
            G=G, intrinsic_dict=intrinsic_dict
        )
        estimated_prob = binary_nested._compute_prob(
            q=np.array(list(result.values())),
            q_indices=q_indices,
            districts=G.districts,
            Ms=Ms,
            Ps=Ps,
        )

        np.testing.assert_array_almost_equal(
            expected_prob, estimated_prob, decimal=2
        )

    def test_that_parameters_of_dag_are_correctly_recovered(self):
        G = dag_chain_small()
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        rng = np.random.default_rng(4213)
        X_1 = rng.binomial(n=1, p=0.5, size=1000)
        X_2 = rng.binomial(n=1, p=expit(1 + X_1))
        df = pd.DataFrame({"X_1": X_1, "X_2": X_2})
        bnm = binary_nested.BinaryNestedModel(graph=G)
        bnm = bnm.fit(X=df)
        results = bnm.fitted_params

        # Check that p(X_2=1 | do(X_1=1)) = 1 - q_{X_2 = 0} (X_1 = 1) = expit(2)
        assert 1 - results[
            (frozenset({"X_2"}), ("X_1",), (1,))
        ] == pytest.approx(expit(2), rel=0.1)
        # Check that p(X_2=1 | do(X_1=0)) = 1 - q_{X_2 = 0} (X_1 = 0) = expit(1)
        assert 1 - results[
            (frozenset({"X_2"}), ("X_1",), (0,))
        ] == pytest.approx(expit(1), rel=0.1)

    def test_estimate_medium_sized_district(self):
        """
        Check that the estimation code works for districts of size 5, based off performance check in Fig 8 of ER13

        :return:
        """
        G = fig8_graph(k=3)
        rng = np.random.default_rng(4213)
        counts = rng.integers(low=60, high=80, size=(2 ** len(G.vertices),))
        expected_prob = counts / counts.sum()
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        bnm = binary_nested.BinaryNestedModel(graph=G)
        bnm = bnm.fit(X=counts)
        q_vector = bnm.fitted_params

        q_indices = binary_nested.compute_q_indices_by_district(
            q_vector_keys=list(q_vector.keys()), districts=G.districts
        )
        Ms, Ps = binary_nested.compute_all_M_and_P(
            G=G, intrinsic_dict=intrinsic_dict
        )
        estimated_prob = binary_nested._compute_prob(
            q=np.array(list(q_vector.values())),
            q_indices=q_indices,
            districts=G.districts,
            Ms=Ms,
            Ps=Ps,
        )

        # Check that the fitted Moebius parameters transform back into the empirical probabilities
        np.testing.assert_array_almost_equal(
            expected_prob, estimated_prob, decimal=2
        )

    def test_estimate_verma_graph(self):
        """
        Check that estimation procedure works for verma graph of size 1

        :return:
        """
        G = verma_graph(k=2)
        rng = np.random.default_rng(4213)
        counts = rng.integers(low=60, high=80, size=(2 ** len(G.vertices),))
        expected_prob = counts / counts.sum()
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()

        bnm = binary_nested.BinaryNestedModel(graph=G)
        bnm = bnm.fit(X=counts)
        q_vector = bnm.fitted_params

        q_indices = binary_nested.compute_q_indices_by_district(
            q_vector_keys=list(q_vector.keys()), districts=G.districts
        )
        Ms, Ps = binary_nested.compute_all_M_and_P(
            G=G, intrinsic_dict=intrinsic_dict
        )
        estimated_prob = binary_nested._compute_prob(
            q=np.array(list(q_vector.values())),
            q_indices=q_indices,
            districts=G.districts,
            Ms=Ms,
            Ps=Ps,
        )

        # Check that the fitted Moebius parameters transform back into the empirical probabilities
        np.testing.assert_array_almost_equal(
            expected_prob, estimated_prob, decimal=2
        )

    def test_synthetic_evans_richardson_fig_8b(self):
        G = wisconsin_fig8b_graph()
        rng = np.random.default_rng(4213)
        U = rng.binomial(n=1, p=0.5, size=1000)
        X = rng.binomial(n=1, p=0.5, size=1000)
        E = rng.binomial(n=1, p=expit(1 + X - U))
        M = rng.binomial(n=1, p=expit(-1 + E))
        Y = rng.binomial(n=1, p=expit(U + X))
        df = pd.DataFrame({"X": X, "E": E, "M": M, "Y": Y})

        # Compute p(Y=1 | X) = \sum_U p(Y=1 | X, U) p(U)
        py1_x1 = 0.5 * expit(2) + 0.5 * expit(1)
        py1_x0 = 0.5 * expit(1) + 0.5 * expit(0)
        bnm = binary_nested.BinaryNestedModel(graph=G)
        bnm = bnm.fit(X=df)
        bnm = bnm.fit(X=df)
        results = bnm.fitted_params
        assert 1 - results[(frozenset({"Y"}), ("X",), (0,))] == pytest.approx(
            py1_x0, rel=0.05
        )
        assert 1 - results[(frozenset({"Y"}), ("X",), (1,))] == pytest.approx(
            py1_x1, rel=0.05
        )

    def test_evans_richardson_fig_8b(self):
        import logging

        binary_nested.logger.setLevel(logging.DEBUG)

        # logging.getLogger("binary_nested").setLevel(logging.WARNING)
        G = wisconsin_fig8b_graph()
        variables = ["X", "E", "M", "Y", "count"]
        values = np.array(
            [
                [0, 0, 0, 0, 241],
                [0, 0, 0, 1, 162],
                [0, 0, 1, 0, 53],
                [0, 0, 1, 1, 39],
                [1, 0, 0, 0, 161],
                [1, 0, 0, 1, 148],
                [1, 0, 1, 0, 33],
                [1, 0, 1, 1, 29],
                [0, 1, 0, 0, 82],
                [0, 1, 0, 1, 176],
                [0, 1, 1, 0, 13],
                [0, 1, 1, 1, 16],
                [1, 1, 0, 0, 113],
                [1, 1, 0, 1, 364],
                [1, 1, 1, 0, 16],
                [1, 1, 1, 1, 30],
            ]
        )
        df = pd.DataFrame(values, columns=variables)

        # try to work out what variables were flipped in the table
        df["X"] = 1 - df["X"]
        df["Y"] = 1 - df["Y"]

        counts = binary_nested.process_data(df, count_variable="count")
        expected_prob = counts / counts.sum()
        bnm = binary_nested.BinaryNestedModel(graph=G)
        bnm = bnm.fit(X=df, tol=1e-8)
        # TODO: It appears that the linear constraint ensuring non-negative probabilities is violated during the optimize call, despite the fact that keep_feasible=True.

        q_vector = bnm.fitted_params
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        q_indices = binary_nested.compute_q_indices_by_district(
            q_vector_keys=list(q_vector.keys()), districts=G.districts
        )
        Ms, Ps = binary_nested.compute_all_M_and_P(
            G=G, intrinsic_dict=intrinsic_dict
        )
        estimated_prob = binary_nested._compute_prob(
            q=np.array(list(q_vector.values())),
            q_indices=q_indices,
            districts=G.districts,
            Ms=Ms,
            Ps=Ps,
        )
        np.testing.assert_allclose(expected_prob, estimated_prob, atol=0.01)

        # Recall that all q params are evaluated at head=0
        # Compute p(Y=1 | do(X)=1) = 1 - q_Y (X=1) ?= 0.5
        assert 1 - bnm.fitted_params[
            (frozenset({"Y"}), ("X",), (1,))
        ] == pytest.approx(0.5, rel=0.01)
        # Compute p(Y=1 | do(X)=1) = 1 - q_Y (X=0) ?= 0.36
        assert 1 - bnm.fitted_params[
            (frozenset({"Y"}), ("X",), (0,))
        ] == pytest.approx(0.36, rel=0.01)

        assert bnm.estimate(
            treatment_dict={"X": 1}, outcome_dict={"Y": 1}
        ) == pytest.approx(0.5, rel=0.01)

        assert bnm.estimate(
            treatment_dict={"X": 0}, outcome_dict={"Y": 1}
        ) == pytest.approx(0.36, rel=0.01)


class TestBinaryNestedModel(unittest.TestCase):
    def test_process_data_on_nonbinary_dataframe_fails(self):
        X = pd.DataFrame({"X1": [0, 0, 2, 1, 1], "X2": [0, 1, 0, 1, 2]})
        with pytest.raises(KeyError):
            result = binary_nested.process_data(X)

    def test_that_process_data_correctly_extracts_counts(self):
        X = pd.DataFrame({"X1": [0, 0, 1, 1, 1], "X2": [0, 1, 0, 1, 1]})
        result = binary_nested.process_data(X)
        expected = np.array([1, 1, 1, 2])
        np.testing.assert_array_equal(result, expected)

        X = pd.DataFrame({"X1": [0, 1, 1, 1, 1], "X2": [0, 1, 0, 1, 1]})
        result = binary_nested.process_data(X)
        expected = np.array([1, 0, 1, 3])
        np.testing.assert_array_equal(result, expected)

    def test_that_process_data_extracts_tabular_summary_correctly(self):
        X = pd.DataFrame(
            {"X2": [0, 0, 1, 1], "X1": [0, 1, 0, 1], "count": [1, 2, 3, 4]}
        )
        result = binary_nested.process_data(X, count_variable="count")
        expected = np.array([1, 3, 2, 4])
        np.testing.assert_array_equal(result, expected)


class TestComputeCounterfactual(unittest.TestCase):
    def test_compute_counterfactual_conditional_ignorability(self):
        G = conditional_ignorability_graph()
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        rng = np.random.default_rng(4213)
        C = rng.binomial(n=1, p=0.5, size=1000)
        A = rng.binomial(n=1, p=expit(1 + C))
        Y = rng.binomial(n=1, p=expit(1 + A + C))
        df = pd.DataFrame({"A": A, "C": C, "Y": Y})
        bnm = binary_nested.BinaryNestedModel(graph=G)
        bnm = bnm.fit(X=df)

        q_vector = bnm.fitted_params

        # p(Y(a=0) = 0)
        x_dict = {"A": 0}
        y_dict = {"Y": 0}

        est_ctf = binary_nested.compute_counterfactual_binary_parameters(
            G, q_vector, x_dict, y_dict
        )

        truth = q_vector[(frozenset({"Y"}), ("A", "C"), (0, 0))] * q_vector[
            (frozenset({"C"}), (), ())
        ] + q_vector[(frozenset({"Y"}), ("A", "C"), (0, 1))] * (
            1 - q_vector[(frozenset({"C"}), (), ())]
        )

        assert truth == pytest.approx(est_ctf)

        # p(Y(a=1) = 1)
        x_dict = {"A": 1}
        y_dict = {"Y": 1}

        est_ctf = binary_nested.compute_counterfactual_binary_parameters(
            G, q_vector, x_dict, y_dict
        )

        truth = (
            1 - q_vector[(frozenset({"Y"}), ("A", "C"), (1, 0))]
        ) * q_vector[(frozenset({"C"}), (), ())] + (
            1 - q_vector[(frozenset({"Y"}), ("A", "C"), (1, 1))]
        ) * (
            1 - q_vector[(frozenset({"C"}), (), ())]
        )

        assert truth == pytest.approx(est_ctf)

    def test_compute_counterfactual_front_door(self):
        df = (
            load_wisconsin_health_study()
            .rename(columns={"E": "A"})
            .groupby(["A", "M", "Y"])["count"]
            .sum()
            .reset_index()
        )

        vertices = ["A", "M", "Y"]
        di_edges = [("A", "M"), ("M", "Y")]
        bi_edges = [("A", "Y")]
        front_door = ADMG(vertices, di_edges, bi_edges)
        bnm = binary_nested.BinaryNestedModel(front_door)
        bnm = bnm.fit(X=df, tol=1e-12)
        theta_M0_A0 = bnm.fitted_params[(frozenset({"M"}), ("A",), (0,))]
        theta_M0_A1 = bnm.fitted_params[(frozenset({"M"}), ("A",), (1,))]
        theta_Y0_M0 = bnm.fitted_params[(frozenset({"Y"}), ("M",), (0,))]
        theta_Y0_M1 = bnm.fitted_params[(frozenset({"Y"}), ("M",), (1,))]
        pY1_A0 = (1 - theta_Y0_M0) * theta_M0_A0 + (1 - theta_Y0_M1) * (
            1 - theta_M0_A0
        )
        pY1_A1 = (1 - theta_Y0_M0) * theta_M0_A1 + (1 - theta_Y0_M1) * (
            1 - theta_M0_A1
        )

        assert pY1_A0 == pytest.approx(bnm.estimate({"A": 0}, {"Y": 1}))
        assert pY1_A1 == pytest.approx(bnm.estimate({"A": 1}, {"Y": 1}))

        assert bnm.fitted_params[
            (frozenset({"Y"}), ("M",), (0,))
        ] == bnm.estimate({"M": 0}, {"Y": 0})

    def test_compute_estimate_without_intervention(self):
        rng = np.random.default_rng(4213)
        U = rng.binomial(n=1, p=0.5, size=1000)
        A = rng.binomial(n=1, p=expit(U), size=1000)
        M = rng.binomial(n=1, p=expit(A), size=1000)
        Y = rng.binomial(n=1, p=expit(1 + M + U))
        vertices = ["A", "M", "Y"]
        di_edges = [("A", "M"), ("M", "Y")]
        bi_edges = [("A", "Y")]
        graph = ADMG(vertices, di_edges, bi_edges)
        df = pd.DataFrame({"A": A, "Y": Y, "M": M})

        bnm = binary_nested.BinaryNestedModel(graph)
        bnm = bnm.fit(X=df)

        assert df["Y"].mean() == pytest.approx(
            bnm.estimate(treatment_dict={}, outcome_dict={"Y": 1})
        )
        assert ((df["Y"] == 1) & (df["A"] == 1)).mean() == pytest.approx(
            bnm.estimate(outcome_dict={"Y": 1, "A": 1})
        )

    def test_compute_estimate_not_id(self):
        rng = np.random.default_rng(4213)
        U = rng.binomial(n=1, p=0.5, size=1000)
        A = rng.binomial(n=1, p=expit(U), size=1000)
        Y = rng.binomial(n=1, p=expit(1 + U))

        df = pd.DataFrame({"A": A, "Y": Y})

        vertices = ["A", "Y"]
        di_edges = [("A", "Y")]
        bi_edges = [("A", "Y")]
        graph = ADMG(vertices, di_edges, bi_edges)
        bnm = binary_nested.BinaryNestedModel(graph)
        bnm = bnm.fit(X=df)

        with pytest.raises(AssertionError) as e:
            result = bnm.estimate(
                outcome_dict={"Y": 1}, treatment_dict={"A": 1}
            )


class TestComputeCounterfactual(unittest.TestCase):
    def test_compute_counterfactual_conditional_ignorability(self):
        G = conditional_ignorability_graph()
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        rng = np.random.default_rng(4213)
        C = rng.binomial(n=1, p=0.5, size=1000)
        A = rng.binomial(n=1, p=expit(1 + C))
        Y = rng.binomial(n=1, p=expit(1 + A + C))
        df = pd.DataFrame({"A": A, "C": C, "Y": Y})
        bnm = binary_nested.BinaryNestedModel(graph=G)
        bnm = bnm.fit(X=df)

        q_vector = bnm.fitted_params

        # p(Y(a=0) = 0)
        x_dict = {"A": 0}
        y_dict = {"Y": 0}

        est_ctf = binary_nested.compute_counterfactual_binary_parameters(
            G, q_vector, x_dict, y_dict
        )

        truth = q_vector[(frozenset({"Y"}), ("A", "C"), (0, 0))] * q_vector[
            (frozenset({"C"}), (), ())
        ] + q_vector[(frozenset({"Y"}), ("A", "C"), (0, 1))] * (
            1 - q_vector[(frozenset({"C"}), (), ())]
        )

        assert truth == pytest.approx(est_ctf)

        # p(Y(a=1) = 1)
        x_dict = {"A": 1}
        y_dict = {"Y": 1}

        est_ctf = binary_nested.compute_counterfactual_binary_parameters(
            G, q_vector, x_dict, y_dict
        )

        truth = (
            1 - q_vector[(frozenset({"Y"}), ("A", "C"), (1, 0))]
        ) * q_vector[(frozenset({"C"}), (), ())] + (
            1 - q_vector[(frozenset({"Y"}), ("A", "C"), (1, 1))]
        ) * (
            1 - q_vector[(frozenset({"C"}), (), ())]
        )

        assert truth == pytest.approx(est_ctf)

    def test_compute_counterfactual_front_door(self):
        df = (
            load_wisconsin_health_study()
            .rename(columns={"E": "A"})
            .groupby(["A", "M", "Y"])["count"]
            .sum()
            .reset_index()
        )

        vertices = ["A", "M", "Y"]
        di_edges = [("A", "M"), ("M", "Y")]
        bi_edges = [("A", "Y")]
        front_door = ADMG(vertices, di_edges, bi_edges)
        bnm = binary_nested.BinaryNestedModel(front_door)
        bnm = bnm.fit(X=df, tol=1e-12)
        theta_M0_A0 = bnm.fitted_params[(frozenset({"M"}), ("A",), (0,))]
        theta_M0_A1 = bnm.fitted_params[(frozenset({"M"}), ("A",), (1,))]
        theta_Y0_M0 = bnm.fitted_params[(frozenset({"Y"}), ("M",), (0,))]
        theta_Y0_M1 = bnm.fitted_params[(frozenset({"Y"}), ("M",), (1,))]
        pY1_A0 = (1 - theta_Y0_M0) * theta_M0_A0 + (1 - theta_Y0_M1) * (
            1 - theta_M0_A0
        )
        pY1_A1 = (1 - theta_Y0_M0) * theta_M0_A1 + (1 - theta_Y0_M1) * (
            1 - theta_M0_A1
        )

        assert pY1_A0 == pytest.approx(bnm.estimate({"A": 0}, {"Y": 1}))
        assert pY1_A1 == pytest.approx(bnm.estimate({"A": 1}, {"Y": 1}))

        assert bnm.fitted_params[
            (frozenset({"Y"}), ("M",), (0,))
        ] == bnm.estimate({"M": 0}, {"Y": 0})

    def test_compute_estimate_without_intervention(self):
        rng = np.random.default_rng(4213)
        U = rng.binomial(n=1, p=0.5, size=1000)
        A = rng.binomial(n=1, p=expit(U), size=1000)
        M = rng.binomial(n=1, p=expit(A), size=1000)
        Y = rng.binomial(n=1, p=expit(1 + M + U))
        vertices = ["A", "M", "Y"]
        di_edges = [("A", "M"), ("M", "Y")]
        bi_edges = [("A", "Y")]
        graph = ADMG(vertices, di_edges, bi_edges)
        df = pd.DataFrame({"A": A, "Y": Y, "M": M})

        bnm = binary_nested.BinaryNestedModel(graph)
        bnm = bnm.fit(X=df)

        assert df["Y"].mean() == pytest.approx(
            bnm.estimate(treatment_dict={}, outcome_dict={"Y": 1})
        )
        assert ((df["Y"] == 1) & (df["A"] == 1)).mean() == pytest.approx(
            bnm.estimate(outcome_dict={"Y": 1, "A": 1})
        )

    def test_compute_estimate_not_id(self):
        rng = np.random.default_rng(4213)
        U = rng.binomial(n=1, p=0.5, size=1000)
        A = rng.binomial(n=1, p=expit(U), size=1000)
        Y = rng.binomial(n=1, p=expit(1 + U))

        df = pd.DataFrame({"A": A, "Y": Y})

        vertices = ["A", "Y"]
        di_edges = [("A", "Y")]
        bi_edges = [("A", "Y")]
        graph = ADMG(vertices, di_edges, bi_edges)
        bnm = binary_nested.BinaryNestedModel(graph)
        bnm = bnm.fit(X=df)

        with pytest.raises(AssertionError) as e:
            result = bnm.estimate(
                outcome_dict={"Y": 1}, treatment_dict={"A": 1}
            )



import copy
import itertools
from typing import Dict, List, Union

import numpy as np
import pgmpy
import sympy as sp
from pgmpy.factors.discrete import TabularCPD

from ..factors import SymCPD
from ..graphs import DAG


class BayesianNetwork(DAG):
    def __init__(self, graph: DAG, cpds: Dict[str, Union[TabularCPD, SymCPD]]):
        """
        A discrete variable Bayesian Network. This is implemented as an augmented
        ananke.graphs.DAG.


        :param graph: A DAG
        :param cpds: A dictionary of vertex names to CPDs. This implementation supports both
        pgmpy's TabularCPD (numpy-backed) as well as ananke's SymCPD (sympy-backed).

        """
        # TODO: raise error if ADMG not DAG

        self.cpds = cpds  # vertex: CPD

        super().__init__(vertices=graph.vertices, di_edges=graph.di_edges)

    def fix(self, variables: List[str]):
        """
        Performs an intervention by setting the conditional distribution of each intervened variable
        to be a point mass at the intervened value. This is a faithful operation (the graph is
        changed and the associated CPD structure reflects the lack of parents).

        """
        super().fix(list(variables))
        if isinstance(variables, dict):
            for vertex, value in variables.items():
                old_cpd = self.cpds[vertex]
                new_values = np.zeros((old_cpd.variable_card, 1))
                new_values[value, :] = 1
                if isinstance(old_cpd.values, np.ndarray):

                    new_cpd = old_cpd.__class__(
                        variable=vertex,
                        variable_card=old_cpd.variable_card,
                        values=np.array(new_values),
                    )
                else:
                    new_cpd = old_cpd.__class__(
                        variable=vertex,
                        variable_card=old_cpd.variable_card,
                        values=old_cpd.values.__class__(new_values),
                    )

                self.cpds[vertex] = new_cpd

        return self

    def copy(self):
        return copy.deepcopy(self)

    def get_cpds(self, vertex):
        return self.cpds[vertex]

    def to_pgmpy(self):
        """
        Converts into pgmpy.models.BayesianNetwork object.
        """
        from pgmpy.models import BayesianNetwork as PgmpyBayesianNetwork

        net = PgmpyBayesianNetwork()
        net.add_nodes_from(self.vertices)
        net.add_edges_from(self.di_edges)

        cpds = [
            x if isinstance(x, TabularCPD) else x.to_pgmpy()
            for x in self.cpds.values()
        ]

        net.add_cpds(*cpds)
        return net


def create_symbolic_cpds(ls_dag, use_uniform_unobs_var=True):
    if hasattr(ls_dag, "get_cards_dict"):
        cards = ls_dag.get_cards_dict()
    else:
        cards = {v.name: v.cardinality for v in ls_dag.vertices.values()}
    if hasattr(ls_dag, "context_variable"):
        context_variable = ls_dag.context_variable
        contexts = ls_dag.contexts
    else:
        context_variable = None
        contexts = None

    all_vars = list()
    cpds = dict()
    for vertex in ls_dag.vertices:
        vertex_vars = []
        param_counter = 1
        relevant_vars = sorted(
            list(ls_dag.parents(vertex))
        )  # some subset of all vars not including v
        fn_map = dict()

        for i, e in enumerate(
            itertools.product(*[range(cards[x]) for x in relevant_vars])
        ):
            for j in range(cards[vertex] - 1):
                # In this if statement, check if intervention occurred
                val_dict = dict(zip(relevant_vars + [vertex], e + (j,)))
                values = e + (j,)
                if use_uniform_unobs_var:
                    if vertex.startswith("U"):
                        fn_map[values] = sp.Rational(1, cards[vertex])
                        continue

                if (
                    context_variable in val_dict
                    and context_variable is not None
                ):
                    current_context = list(contexts)[val_dict[context_variable]]
                    if vertex in current_context:
                        current_ix = current_context.index(vertex)
                        prob = list(contexts.values())[
                            val_dict[context_variable]
                        ][current_ix][j]
                        fn_map[values] = sp.Rational(prob)
                        continue
                if e:

                    symbol = sp.Symbol(
                        f"q_{vertex}_{j}_{''.join([str(x) for x in e])}"
                    )
                else:
                    symbol = sp.Symbol(f"q_{vertex}_{j}")

                fn_map[values] = symbol
                all_vars.append(symbol)
                param_counter += 1

        def func(
            vars,
            relevant_vars=relevant_vars,
            fn_map=fn_map,
            func_var=vertex,
            cards=cards,
        ):
            fn_tuple = tuple([vars[x] for x in relevant_vars + [func_var]])

            if vars[func_var] == cards[func_var] - 1:

                all_fn_tuples = [
                    tuple(
                        [
                            vars[x] if x != func_var else i
                            for x in relevant_vars + [func_var]
                        ]
                    )
                    for i in range(cards[func_var] - 1)
                ]
                return 1 - sum([fn_map[tup] for tup in all_fn_tuples])

            else:
                return fn_map[fn_tuple]

        final_values = sp.Array(
            [
                [
                    func(dict(zip(relevant_vars, e)) | {vertex: i})
                    for e in itertools.product(
                        *[range(cards[x]) for x in relevant_vars]
                    )
                ]
                for i in range(cards[vertex])
            ]
        )
        if relevant_vars:
            cpd = SymCPD(
                variable=vertex,
                variable_card=cards[vertex],
                values=final_values,
                evidence=relevant_vars,
                evidence_card=[cards[v] for v in relevant_vars],
            )
        else:
            cpd = SymCPD(
                variable=vertex,
                variable_card=cards[vertex],
                values=final_values,
            )

        cpds[vertex] = cpd

    return cpds, all_vars


def generate_random_cpds(graph, dir_conc=10, context_variable="S", seed=42):
    """
    Given a graph and a set of cardinalities for variables in a DAG, constructs random conditional probability distributions. Supports optional contexts and context variable to generate CPDs consistent with a context specific DAG for data fusion.

    :param graph: A graph whose variables have cardinalities, and optionally
    :type Graph: DAG, ADMG, LSADMG
    :param dir_conc: The Dirichlet concetration parameter
    :param context_variable: Name of the context variable
    """
    rng = np.random.default_rng(seed=seed)

    cpds = dict()
    cards = {v: graph.vertices[v].cardinality for v in graph.vertices}
    for k, v in cards.items():
        if v is None:
            raise ValueError(
                "Invalid cardinality provided for vertex {}".format(k)
            )
    if hasattr(graph, "contexts"):
        contexts = list(graph.contexts.keys())
        context_distributions = list(graph.contexts.values())
    else:
        contexts = None

    for vertex in sorted(graph.vertices):
        parents = sorted(list(graph.parents(vertex)))

        if context_variable not in parents or contexts is None:
            if not graph.parents(vertex):
                values = rng.dirichlet(
                    cards[vertex] * [dir_conc],
                    1,
                ).T
                cpd = TabularCPD(
                    variable=vertex, variable_card=cards[vertex], values=values
                )

            else:
                values = rng.dirichlet(
                    cards[vertex] * [dir_conc],
                    np.prod([cards[x] for x in parents]),
                ).T

                cpd = TabularCPD(
                    variable=vertex,
                    variable_card=cards[vertex],
                    values=values,
                    evidence=parents,
                    evidence_card=[cards[x] for x in parents],
                )

        else:
            no_s_parents = parents.copy()
            no_s_parents.remove(context_variable)
            no_s_parents = list(no_s_parents)
            reordered_parents = [context_variable] + no_s_parents
            s_specific_values = rng.dirichlet(
                cards[vertex] * [dir_conc],
                int(np.prod([cards[x] for x in no_s_parents])),
            ).T
            values = []
            for i, context in enumerate(contexts):
                if vertex in set(context):
                    ix = context.index(vertex)
                    distribution = context_distributions[i][ix]
                    intervened_values = np.tile(
                        distribution,
                        (np.prod([cards[x] for x in no_s_parents]), 1),
                    ).T
                    values.append(intervened_values)
                else:
                    values.append(s_specific_values)
            values = np.hstack(values)
            cpd = TabularCPD(
                variable=vertex,
                variable_card=cards[vertex],
                values=values,
                evidence=reordered_parents,
                evidence_card=[cards[x] for x in reordered_parents],
            )

        cpds[vertex] = cpd

    return cpds


def intervene(net, treatment_dict):
    """
    Performs an intervention on a pgmpy.models.BayesianNetwork, by setting the conditional distribution of
    each intervened variable to be a point mass at the intervened value. Does not alter the structure of the parents of the network (i.e. is a non-faithful operation).

    If you have an ananke.models.BayesianNetwork, consider using the .fix(treatment_dict) method
    instead, which has the further advantage of performing the operation faithfully (the underlying
    DAG is modified accordingly, and the parents of the intervened variables
    in that conditional probability distributions are removed).


    :param net: pgmpy.models.Bayesian Network
    :type net: pgmpy.models.BayesianNetwork
    :param treatment_dict: dictionary of variables to values:
    :type treatment_dict: dict
    """
    net_copy = net.copy()

    for vertex, value in treatment_dict.items():
        old_cpd = net_copy.get_cpds(vertex)
        old_values = old_cpd.get_values()
        new_values = np.zeros(old_values.shape)
        new_values[value, :] = 1
        new_cpd = TabularCPD(
            variable=vertex,
            variable_card=old_cpd.variable_card,
            values=new_values,
            evidence=old_cpd.variables[1:],
            evidence_card=old_cpd.cardinality[1:],
        )
        net_copy.add_cpds(new_cpd)

    return net_copy



"""
Class for missing ID
"""


class MissingFullID:
    def __init__(self, graph):
        """
        Constructor

        :param graph: Missing data graph to run the ID algorithm on
        """

        self.graph = graph

    def id(self):
        """
        Function to ID the full law

        :return: boolean is ID or not
        """

        for Ri in self.graph.vertices:

            # Only look at the missingness indicators in the list of vertices
            if not Ri.startswith("R_"):
                continue

            # Find children of Ri that are not in proxy set Xp
            child_Ri = set(
                [v for v in self.graph.children([Ri]) if v.startswith("R_")]
            )

            # Find Markov blanket of the set {Ri, child_Ri} and add back child_Ri to the set
            colluding_path_Ri = self.graph.markov_blanket(
                set.union({Ri}, child_Ri)
            ).union(child_Ri)

            # Find Xi (the counterfactual X with the corresponding missingness indicator Ri)
            Xi = "X_" + Ri.split("_")[1]

            # If Xi is in colluding_path_Ri, then there is a colluding path between Ri and Xi.
            # Therefore, full law is not ID.
            if Xi in colluding_path_Ri:
                return False

        return True



import unittest

from ananke.estimation import AutomatedIF
from ananke.graphs import ADMG


class TestADMG(unittest.TestCase):
    def test_conditionally_ignorable_model(self):

        vertices = ["A", "C", "Y"]
        di_edges = [("C", "A"), ("C", "Y"), ("A", "Y")]
        G = ADMG(vertices, di_edges)
        influence = AutomatedIF(G, "A", "Y")
        print(influence.beta_primal_)
        print(influence.beta_dual_)
        print(influence.nonparametric_if_)
        print(influence.eff_if_)
        self.assertEqual(influence.eff_if_, influence.nonparametric_if_)

    def test_mbshielded_afixable(self):

        vertices = ["Z1", "Z2", "C1", "C2", "T", "M", "Y"]
        di_edges = [
            ("C1", "Z1"),
            ("C1", "T"),
            ("C1", "M"),
            ("C2", "Z1"),
            ("C2", "T"),
            ("C2", "M"),
            ("Z1", "Z2"),
            ("Z2", "T"),
            ("T", "M"),
            ("M", "Y"),
        ]
        bi_edges = [("Z1", "T"), ("Z2", "C1"), ("C2", "Y")]
        G = ADMG(vertices, di_edges, bi_edges)
        influence = AutomatedIF(G, "T", "Y")
        print(influence.beta_primal_)
        print(influence.beta_dual_)
        print(influence.nonparametric_if_)
        print(influence.eff_if_)
        self.assertNotEqual(influence.eff_if_, influence.nonparametric_if_)

    def test_frontdoor_model(self):

        vertices = ["A", "M", "Y"]
        di_edges = [("A", "M"), ("M", "Y")]
        bi_edges = [("A", "Y")]
        G = ADMG(vertices, di_edges, bi_edges)
        influence = AutomatedIF(G, "A", "Y")
        print(influence.beta_primal_)
        print(influence.beta_dual_)
        print(influence.nonparametric_if_)
        print(influence.eff_if_)
        self.assertEqual(influence.eff_if_, influence.nonparametric_if_)

    def test_mbshielded_pfixable(self):

        vertices = ["C1", "C2", "Z1", "Z2", "T", "M", "L", "Y"]
        di_edges = [
            ("C1", "T"),
            ("C1", "L"),
            ("C2", "T"),
            ("C2", "M"),
            ("C2", "L"),
            ("C2", "Y"),
            ("T", "M"),
            ("M", "L"),
            ("L", "Y"),
        ]
        bi_edges = [("Z1", "C1"), ("Z2", "C2"), ("T", "L")]
        G = ADMG(vertices, di_edges, bi_edges)
        influence = AutomatedIF(G, "T", "Y")
        print(influence.beta_primal_)
        print(influence.beta_dual_)
        print(influence.nonparametric_if_)
        print(influence.eff_if_)
        self.assertNotEqual(influence.eff_if_, influence.nonparametric_if_)

    def test_verma_afixable(self):
        vertices = ["T", "M", "L", "Y"]
        di_edges = [("T", "M"), ("M", "L"), ("L", "Y")]
        bi_edges = [("M", "Y")]
        G = ADMG(vertices, di_edges, bi_edges)
        influence = AutomatedIF(G, "T", "Y")
        print(influence.beta_primal_)
        print(influence.beta_dual_)
        print(influence.nonparametric_if_)
        print(influence.eff_if_)
        self.assertEqual(
            influence.eff_if_, "Cannot compute, graph is not mb-shielded."
        )

    def test_verma_pfixable(self):
        vertices = ["T", "M", "L", "Y"]
        di_edges = [("T", "M"), ("M", "L"), ("L", "Y")]
        bi_edges = [("M", "Y"), ("T", "L")]
        G = ADMG(vertices, di_edges, bi_edges)
        influence = AutomatedIF(G, "T", "Y")
        print(influence.beta_primal_)
        print(influence.beta_dual_)
        print(influence.nonparametric_if_)
        print(influence.eff_if_)
        self.assertEqual(
            influence.eff_if_, "Cannot compute, graph is not mb-shielded."
        )

    def test_nested_fixable_and_notid(self):

        G_bowarc = ADMG(
            vertices=["T", "Y"], di_edges=[("T", "Y")], bi_edges=[("T", "Y")]
        )
        with self.assertRaises(RuntimeError):
            AutomatedIF(G_bowarc, "T", "Y")
        vertices = ["C1", "C2", "T", "M", "Z", "R1", "R2", "Y"]
        di_edges = [
            ("C1", "T"),
            ("C1", "Y"),
            ("C2", "T"),
            ("C2", "Y"),
            ("R2", "Y"),
            ("Z", "T"),
            ("T", "R1"),
            ("T", "Y"),
            ("R1", "M"),
            ("M", "Y"),
        ]
        bi_edges = [
            ("Z", "R2"),
            ("T", "R2"),
            ("Z", "R1"),
            ("C1", "M"),
            ("C1", "Y"),
            ("C2", "M"),
            ("C2", "Y"),
        ]
        G_nfixable = ADMG(vertices, di_edges, bi_edges)
        with self.assertRaises(NotImplementedError):
            AutomatedIF(G_nfixable, "T", "Y")


if __name__ == "__main__":
    unittest.main()



import functools

import sympy as sp

from ananke.factors import SymCPD
from ananke.graphs import DAG
from ananke.inference.variable_elimination import variable_elimination
from ananke.models import BayesianNetwork


def test_variable_elimination_on_three_var_chain():
    x1 = sp.Symbol("x1")
    y1 = sp.Symbol("y1")
    y2 = sp.Symbol("y2")
    x2 = sp.Symbol("x2")
    x3 = sp.Symbol("x3")
    cpd_a = SymCPD("A", 2, sp.Array([[x1, y1], [1 - x1, 1 - y1]]), ["B"], [2])
    cpd_b = SymCPD("B", 2, sp.Array([[x2, y2], [1 - x2, 1 - y2]]), ["C"], [2])
    cpd_c = SymCPD("C", 2, sp.Array([[x3], [1 - x3]]))

    graph = DAG(
        vertices={"A": 2, "B": 2, "C": 2},
        di_edges=[("C", "B"), ("B", "A")],
    )
    cpds = {"A": cpd_a, "B": cpd_b, "C": cpd_c}
    bayesian_network = BayesianNetwork(graph, cpds)
    marginal = variable_elimination(bayesian_network, variables=["A"])
    truth = functools.reduce(
        lambda x, y: x.product(y), [cpd.to_factor() for cpd in cpds.values()]
    ).marginalize(["B", "C"])
    assert sp.simplify(sp.tensorcontraction(marginal.values, (0,))) == 1
    assert marginal == truth



import unittest

from ananke.graphs import ADMG, MissingADMG
from ananke.identification import MissingFullID


class TestMissingID(unittest.TestCase):
    def test_full_id(self):
        vertices = ["X_1", "X_2", "R_1", "R_2"]
        di_edges = [("X_1", "X_2"), ("X_1", "R_2"), ("R_2", "R_1")]
        mdag = ADMG(vertices, di_edges)
        full_id = MissingFullID(mdag)
        self.assertTrue(full_id.id())

    def test_full_id_colluder(self):
        vertices = ["X_1", "X_2", "R_1", "R_2"]
        di_edges = [("X_1", "R_2"), ("R_1", "R_2")]
        mdag = ADMG(vertices, di_edges)
        full_id = MissingFullID(mdag)
        self.assertFalse(full_id.id())

    def test_full_id_cross_pattern(self):

        vertices = ["X_1", "X_2", "R_1", "R_2"]
        di_edges = [("X_1", "R_2"), ("X_2", "R_1"), ("R_2", "R_1")]
        mdag = ADMG(vertices, di_edges)
        full_id = MissingFullID(mdag)
        self.assertFalse(full_id.id())

    def test_full_id_chain(self):

        vertices = ["X_1", "X_2", "X_3", "R_1", "R_2", "R_3"]
        di_edges = [
            ("X_1", "R_2"),
            ("X_2", "R_3"),
            ("X_3", "R_1"),
            ("R_3", "R_2"),
            ("R_2", "R_1"),
        ]
        mdag = ADMG(vertices, di_edges)
        full_id = MissingFullID(mdag)
        self.assertTrue(full_id.id())
        mdag.add_diedge("X_1", "R_3")
        full_id = MissingFullID(mdag)
        self.assertTrue(full_id.id())

    def test_full_id_complicated(self):

        vertices = [
            "X_1",
            "X_2",
            "X_3",
            "X_4",
            "X_5",
            "R_1",
            "R_2",
            "R_3",
            "R_4",
            "R_5",
        ]
        di_edges = [
            ("X_1", "R_2"),
            ("X_2", "R_3"),
            ("X_2", "R_4"),
            ("X_2", "R_5"),
            ("X_3", "R_1"),
            ("X_4", "R_3"),
            ("X_4", "R_5"),
            ("X_5", "R_3"),
            ("X_5", "R_4"),
            ("R_3", "R_2"),
            ("R_3", "R_4"),
            ("R_3", "R_5"),
            ("R_4", "R_2"),
            ("R_5", "R_2"),
        ]
        mdag = ADMG(vertices, di_edges)
        full_id = MissingFullID(mdag)
        self.assertTrue(full_id.id())
        mdag.add_diedge("R_2", "R_1")
        full_id = MissingFullID(mdag)
        self.assertTrue(full_id.id())

    def test_full_not_ID(self):

        vertices = [
            "X_1",
            "X_2",
            "X_3",
            "R_1",
            "R_2",
            "R_3",
            "Xp_1",
            "Xp_2",
            "Xp_3",
        ]
        di_edges = [
            ("X_1", "X_2"),
            ("X_1", "R_2"),
            ("X_1", "R_3"),
            ("X_2", "R_3"),
            ("X_2", "X_3"),
            ("X_3", "R_1"),
            ("R_3", "R_2"),
            ("R_2", "R_1"),
        ]
        bi_edges = [("X_1", "X_3"), ("X_2", "R_3")]
        mdag = MissingADMG(vertices, di_edges, bi_edges)
        full_id = MissingFullID(mdag)
        self.assertTrue(full_id.id)
        mdag.add_biedge("R_1", "R_3")
        full_id = MissingFullID(mdag)
        self.assertFalse(full_id.id())


if __name__ == "__main__":
    unittest.main()



"""
The code relies on the following papers:

[ER12] Evans, R. J., & Richardson, T. S. (2012). Maximum likelihood fitting of acyclic directed mixed graphs to binary data. arXiv preprint arXiv:1203.3479.
[ER19] Evans, R. J., & Richardson, T. S. (2019). Smooth, identifiable supermodels of discrete DAG models with latent variables. Bernoulli, 25(2), 848-876.

"""
import unittest
from collections import OrderedDict

import numpy as np
import pandas as pd
import pytest
from scipy.special import expit

from ananke.datasets import load_wisconsin_health_study
from ananke.graphs import ADMG, DAG
from ananke.models import binary_nested


def dag_chain_small():
    vertices = ["X_1", "X_2"]
    di_edges = [("X_1", "X_2")]
    G = DAG(di_edges=di_edges, vertices=vertices)

    return G


def wisconsin_fig8b_graph():
    G = ADMG(
        vertices=["X", "E", "M", "Y"],
        di_edges=[("X", "E"), ("E", "M"), ("X", "Y")],
        bi_edges=[("E", "Y")],
    )

    return G


def sixteen_not_eighteen_graph():
    vertices = ["A", "B", "C", "D", "E"]
    di_edges = [("A", "B"), ("B", "C"), ("C", "D")]
    bi_edges = [("E", "D"), ("B", "E")]

    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def standard_graph():
    """
    Fig 1 from [ER12]
    :return:
    """
    vertices = ["X1", "X2", "X3", "X4"]
    di_edges = [("X1", "X2"), ("X2", "X4")]
    bi_edges = [("X3", "X2"), ("X3", "X4")]
    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def standard_dag():
    vertices = ["X1", "X2", "X3", "X4"]
    di_edges = [("X1", "X2"), ("X2", "X3"), ("X3", "X4")]
    bi_edges = list()
    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def fig1b_graph():
    """
     Fig 1b from [ER19]

    :return:
    """
    vertices = ["X", "E", "M", "Y"]
    di_edges = [("X", "E"), ("E", "M"), ("M", "Y")]
    bi_edges = [("E", "Y")]
    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def fig2_graph():
    """
    Fig 2 from [ER19]

    :return:
    """
    vertices = ["X1", "X2", "X3", "X4"]
    di_edges = [("X1", "X3"), ("X2", "X4")]
    bi_edges = [("X1", "X4"), ("X2", "X3")]
    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def fig3_graph():
    """
    Fig 3 from [ER12]

    :return:
    """
    vertices = ["X1", "X2", "X3"]
    di_edges = [("X1", "X2")]
    bi_edges = [("X3", "X2"), ("X1", "X3")]
    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def fig7_graph():
    """
    Fig 7 from [ER19]
    :return:
    """
    vertices = ["2", "3", "4", "5"]
    di_edges = [("2", "3"), ("3", "4")]
    bi_edges = [("2", "5"), ("5", "4")]
    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def fig8_graph(k=5):
    """
    Fig 8 from [ER12]
    :param k:
    :return:
    """
    vertices = ["X{}".format(i) for i in range(k)] + [
        "Y{}".format(i) for i in range(k)
    ]
    di_edges = [("X{}".format(i), "Y{}".format(i)) for i in range(k)]
    bi_edges = [
        ("Y{}".format(i), "Y{}".format(i + 1)) for i in range(k - 1)
    ] + [("X{}".format(i), "X{}".format(i + 1)) for i in range(k - 1)]
    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def verma_graph(k=5):
    """
    Non-ancestral graph containing Verma constraints
    :param k:
    :return:
    """
    vertices = (
        ["A", "Y"]
        + ["L{}".format(i) for i in range(k)]
        + ["B{}".format(i) for i in range(k)]
    )
    di_edges = (
        [("A", "L0")]
        + [("L{}".format(i), "B{}".format(i)) for i in range(k)]
        + [("B{}".format(i), "Y") for i in range(k)]
    )
    bi_edges = [("L{}".format(i), "Y") for i in range(k)] + [
        ("B{}".format(i), "L{}".format(i + 1)) for i in range(k - 1)
    ]

    G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)

    return G


def conditional_ignorability_graph():

    vertices = ["A", "C", "Y"]
    di_edges = [("C", "A"), ("C", "Y"), ("A", "Y")]
    G = DAG(di_edges=di_edges, vertices=vertices)

    return G


class TestBinaryNested(unittest.TestCase):
    def test_permutation(self):
        """
        Check that permutation helper function is correct.

        :return:
        """
        result = binary_nested.permutations(2, k=2)
        expected = [(0, 0), (0, 1), (1, 0), (1, 1)]
        self.assertEqual(expected, result)
        result = binary_nested.permutations(2, k=3)
        expected = [
            (0, 0),
            (0, 1),
            (0, 2),
            (1, 0),
            (1, 1),
            (1, 2),
            (2, 0),
            (2, 1),
            (2, 2),
        ]
        self.assertEqual(expected, result)

    def test_recursive_partition_heads(self):
        """
        Check that recursive partition heads are correctly formed.

        :return:
        """
        G_1 = standard_graph()
        intrinsic_dict, fixing_orders = G_1.get_intrinsic_sets_and_heads()
        heads = binary_nested.recursive_partition_heads(
            {"X1", "X2"}, intrinsic_dict
        )
        self.assertEqual({frozenset({"X2"}), frozenset({"X1"})}, heads)
        heads = binary_nested.recursive_partition_heads(
            {"X3", "X4"}, intrinsic_dict
        )
        self.assertEqual({frozenset({"X3", "X4"})}, heads)

        G_2 = fig1b_graph()
        intrinsic_dict, fixing_orders = G_2.get_intrinsic_sets_and_heads()
        heads = binary_nested.recursive_partition_heads(
            {"X", "E", "Y"}, intrinsic_dict
        )
        self.assertEqual({frozenset({"X"}), frozenset({"E", "Y"})}, heads)
        heads = binary_nested.recursive_partition_heads(
            {"M", "Y"}, intrinsic_dict
        )
        self.assertEqual({frozenset({"M"}), frozenset({"Y"})}, heads)

    def test_maximal_heads(self):
        """
        Check that maximal heads function correctly extracts the maximal head
        :return:
        """
        G = fig1b_graph()
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        result = binary_nested.maximal_heads(
            {frozenset({"E"}), frozenset({"E", "Y"})}, intrinsic_dict
        )
        self.assertEqual([frozenset({"Y", "E"})], result)

    def test_q_vector(self):
        """
        Test that q_vector initializes parameters with keys in the expected order. Also checks that the values are
        initialized to represent full independence.

        :return:
        """
        G = fig1b_graph()
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        result = binary_nested.initialize_q_vector(intrinsic_dict)
        self.assertIn((frozenset({"E", "Y"}), ("M", "X"), (0, 1)), result)
        self.assertNotIn((frozenset({"E", "Y"}), ("X", "M"), (1, 0)), result)
        expected_keys = [
            (frozenset({"E"}), ("X",), (0,)),
            (frozenset({"E"}), ("X",), (1,)),
            (frozenset({"E", "Y"}), ("M", "X"), (0, 0)),
            (frozenset({"E", "Y"}), ("M", "X"), (0, 1)),
            (frozenset({"E", "Y"}), ("M", "X"), (1, 0)),
            (frozenset({"E", "Y"}), ("M", "X"), (1, 1)),
            (frozenset({"M"}), ("E",), (0,)),
            (frozenset({"M"}), ("E",), (1,)),
            (frozenset({"X"}), (), ()),
            (frozenset({"Y"}), ("M",), (0,)),
            (frozenset({"Y"}), ("M",), (1,)),
        ]

        self.assertEqual(expected_keys, list(result.keys()))

        self.assertAlmostEqual(
            result[(frozenset({"E", "Y"}), ("M", "X"), (1, 1))], 0.25
        )
        self.assertAlmostEqual(result[(frozenset({"Y"}), ("M",), (1,))], 0.5)

    def test_heads_tails_map(self):
        G = fig1b_graph()
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        result = binary_nested.get_heads_tails_map(intrinsic_dict)
        self.assertEqual(result[frozenset({"Y", "E"})], ("M", "X"))

    def test_compute_likelihood_district(self):
        """
        Example 1 from [ER12]

        :return:
        """
        G = standard_graph()
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        q_vector = binary_nested.initialize_q_vector(intrinsic_dict)
        heads_tails_map = binary_nested.get_heads_tails_map(
            intrinsic_dict=intrinsic_dict
        )

        nu_dict = {"X1": 1, "X2": 1, "X3": 0, "X4": 1}
        lld_value = 1
        for D_j in G.districts:
            # send nu_dict as a dictionary : if you dont want the input that way, include a line to map vector nu to the values in nu_dict
            # nu_dict always contains all vertices V - don't change according to district
            lld_value *= binary_nested.compute_likelihood_district(
                nu_dict, q_vector, D_j, heads_tails_map, intrinsic_dict
            )

        expected = (1 - q_vector[(frozenset({"X1"}), (), ())]) * (
            q_vector[(frozenset({"X3"}), (), ())]
            - q_vector[(frozenset({"X2", "X3"}), ("X1",), (1,))]
            - q_vector[(frozenset({"X3", "X4"}), ("X1", "X2"), (1, 1))]
            + q_vector[(frozenset({"X3", "X4"}), ("X1", "X2"), (1, 1))]
            * q_vector[(frozenset({"X2"}), ("X1",), (1,))]
        )

        self.assertAlmostEqual(expected, lld_value)

    def test_compute_terms(self):
        """
        Check that the terms (product of various q parameters) are computed correctly. These are represented by tuples
        consisting of (all heads, all tails, value of all tails).

        :return:
        """
        G = fig3_graph()
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        heads_tails_map = binary_nested.get_heads_tails_map(
            intrinsic_dict=intrinsic_dict
        )
        for D_j in G.districts:
            result = binary_nested.compute_terms(
                D_j, intrinsic_dict, heads_tails_map
            )

            expected = [
                (frozenset(), (), ()),
                (frozenset({"X1"}), (), ()),
                (frozenset({"X2"}), ("X1",), (0,)),
                (frozenset({"X2"}), ("X1",), (1,)),
                (frozenset({"X2", "X1"}), ("X1",), (0,)),
                (frozenset({"X2", "X1"}), ("X1",), (1,)),
                (frozenset({"X3"}), (), ()),
                (frozenset({"X3", "X1"}), (), ()),
                (frozenset({"X2", "X3"}), ("X1",), (0,)),
                (frozenset({"X2", "X3"}), ("X1",), (1,)),
                (frozenset({"X2", "X3", "X1"}), ("X1",), (0,)),
                (frozenset({"X2", "X3", "X1"}), ("X1",), (1,)),
            ]
            self.assertEqual(expected, result)

    def test_compute_M(self):
        """
        Check form of M matrix as shown in [ER12] is correct for provided example graph

        :return:
        """
        G = fig3_graph()
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        heads_tails_map = binary_nested.get_heads_tails_map(
            intrinsic_dict=intrinsic_dict
        )
        for D_j in G.districts:
            terms = binary_nested.compute_terms(
                D_j, intrinsic_dict, heads_tails_map
            )
            partition_head_dict = binary_nested.compute_partition_head_dict(
                intrinsic_dict=intrinsic_dict, district=D_j
            )
            result = binary_nested.compute_M(
                G=G,
                partition_head_dict=partition_head_dict,
                district=D_j,
                heads_tails_map=heads_tails_map,
                terms=terms,
            )
        ix = binary_nested.permutations(3).index((1, 0, 1))
        expected = OrderedDict(
            [
                ((frozenset(), (), ()), 0),
                ((frozenset({"X1"}), (), ()), 0),
                ((frozenset({"X2"}), ("X1",), (0,)), 0),
                ((frozenset({"X2"}), ("X1",), (1,)), 1),
                ((frozenset({"X2", "X1"}), ("X1",), (0,)), 0),
                ((frozenset({"X2", "X1"}), ("X1",), (1,)), -1),
                ((frozenset({"X3"}), (), ()), 0),
                ((frozenset({"X3", "X1"}), (), ()), 0),
                ((frozenset({"X3", "X2"}), ("X1",), (0,)), 0),
                ((frozenset({"X3", "X2"}), ("X1",), (1,)), -1),
                ((frozenset({"X2", "X3", "X1"}), ("X1",), (0,)), 0),
                ((frozenset({"X2", "X3", "X1"}), ("X1",), (1,)), 1),
            ]
        )

        self.assertEqual(list(expected.values()), result[ix, :].tolist())
        self.assertEqual(result.shape, (2 ** len(G.vertices), len(terms)))

    def test_compute_P(self):
        """
        Check form of P matrix as shown in [ER12] is correct for provided example graph

        :return:
        """
        G = fig3_graph()
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        heads_tails_map = binary_nested.get_heads_tails_map(
            intrinsic_dict=intrinsic_dict
        )
        q_vector = binary_nested.initialize_q_vector(intrinsic_dict)
        # Loops over the whole graph which is a single district
        for D_j in G.districts:
            terms = binary_nested.compute_terms(
                D_j, intrinsic_dict, heads_tails_map
            )
            partition_head_dict = binary_nested.compute_partition_head_dict(
                intrinsic_dict=intrinsic_dict, district=D_j
            )
            result = binary_nested.compute_P(
                partition_head_dict=partition_head_dict,
                q_vector_keys=q_vector,
                heads_tails_map=heads_tails_map,
                terms=terms,
            )

            expected = [1, 0, 0, 0, 0, 1, 0]
            ix = terms.index((frozenset({"X1", "X2", "X3"}), ("X1",), (1,)))
            self.assertEqual(result[ix, :].tolist(), expected)
            self.assertEqual(result.shape, (len(terms), len(q_vector)))

    def test_compute_P_Q_sums_to_1(self):
        """
        Check that the property that probabilities obtained after transformation of q parameters sum to one.
        Note that this property is guaranteed by the correct form of the transformation, and applies to all choices of
        q. However, it is not guaranteed that the probabilities are non-negative.

        :return:
        """
        G = fig2_graph()
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        heads_tails_map = binary_nested.get_heads_tails_map(
            intrinsic_dict=intrinsic_dict
        )
        q_vector = binary_nested.initialize_q_vector(intrinsic_dict)
        prob = 1
        for D_j in G.districts:
            terms = binary_nested.compute_terms(
                D_j, intrinsic_dict, heads_tails_map
            )
            partition_head_dict = binary_nested.compute_partition_head_dict(
                intrinsic_dict=intrinsic_dict, district=D_j
            )
            M = binary_nested.compute_M(
                G=G,
                partition_head_dict=partition_head_dict,
                district=D_j,
                heads_tails_map=heads_tails_map,
                terms=terms,
            )
            P = binary_nested.compute_P(
                partition_head_dict=partition_head_dict,
                q_vector_keys=q_vector,
                heads_tails_map=heads_tails_map,
                terms=terms,
            )

            self.assertEqual(M.shape, (2 ** len(G.vertices), len(terms)))
            self.assertEqual(P.shape, (len(terms), len(q_vector)))

            prob *= M @ np.exp(P @ np.log(np.array(list(q_vector.values()))))

        # Initialized q_vector is not guaranteed to satisfy constraints (and thus be within [0,1]), but vector of probs will sum to 1
        self.assertAlmostEqual(1, np.sum(prob))

    def test_a_y1_y2_graph(self):
        """
        Check that q(A) is a parameter prior to fixing, but not after.

        :return:
        """
        vertices = ["A", "Y1", "Y2"]
        di_edges = [("A", "Y1")]
        bi_edges = [("Y1", "Y2"), ("A", "Y1")]
        G = ADMG(di_edges=di_edges, bi_edges=bi_edges, vertices=vertices)
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()

        q_vector = binary_nested.initialize_q_vector(intrinsic_dict)

        self.assertTrue((frozenset({"A"}), (), ()) in q_vector)

        G.fix("A")

        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()

        q_vector = binary_nested.initialize_q_vector(intrinsic_dict)
        self.assertTrue((frozenset({"A"}), (), ()) not in q_vector)

    def test_compute_prob(self):
        """

        :return:
        """
        G = fig2_graph()
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        q_vector = binary_nested.initialize_q_vector(intrinsic_dict)
        districts = G.districts
        Ms, Ps = binary_nested.compute_all_M_and_P(
            G=G, intrinsic_dict=intrinsic_dict
        )

        q_indices = binary_nested.compute_q_indices_by_district(
            q_vector_keys=list(q_vector.keys()), districts=districts
        )

        prob = binary_nested._compute_prob(
            q=np.array(list(q_vector.values())),
            q_indices=q_indices,
            districts=districts,
            Ms=Ms,
            Ps=Ps,
        )
        self.assertAlmostEqual(1, np.sum(prob))

    def test_compute_q_constraint_from_A_b_matrices(self):
        """
        Test that the A and b matrices are constructed correctly

        :return:
        """
        G = fig3_graph()
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        heads_tails_map = binary_nested.get_heads_tails_map(
            intrinsic_dict=intrinsic_dict
        )
        q_vector = binary_nested.initialize_q_vector(intrinsic_dict)
        for D_j in G.districts:
            terms = binary_nested.compute_terms(
                D_j, intrinsic_dict, heads_tails_map
            )

            partition_head_dict = binary_nested.compute_partition_head_dict(
                intrinsic_dict=intrinsic_dict, district=D_j
            )

            M = binary_nested.compute_M(
                G=G,
                partition_head_dict=partition_head_dict,
                district=D_j,
                heads_tails_map=heads_tails_map,
                terms=terms,
            )
            # Select q parameters whose heads are in district D_j
            new_q = OrderedDict(
                [(k, v) for k, v in q_vector.items() if set(k[0]).issubset(D_j)]
            )
            P = binary_nested.compute_P(
                partition_head_dict=partition_head_dict,
                q_vector_keys=new_q,
                heads_tails_map=heads_tails_map,
                terms=terms,
            )

            variable = "X2"
            variables = list(G.vertices)
            keys = list(q_vector.keys())
            theta_bool_map = {
                v: np.array(
                    [
                        True if v in item[0] else False
                        for i, item in enumerate(keys)
                    ]
                )
                for v in variables
            }
            A, b = binary_nested.construct_A_b(
                variable=variable,
                q=np.array(list(q_vector.values())),
                theta_reindexed_bool_map=theta_bool_map,
                M=M,
                P=P,
            )

            # We check the row 5 corresponding to state 101.
            # This should have A[5, :] = [0, 1-q_1, 0, -1 + q_1] and b[5] = 0

            np.testing.assert_array_almost_equal(
                A[5, :],
                np.array(
                    [
                        0,
                        1 - q_vector[(frozenset({"X1"}), (), ())],
                        0,
                        -1 + q_vector[(frozenset({"X1"}), (), ())],
                    ]
                ),
            )

            self.assertAlmostEqual(b[5], 0)

    def test_that_p_align_from_partial_likelihood_and_prob(self):
        """
        Checks that the probabilities obtained from the full probability function aligns with the method used to compute
        partial probabilities, at the same q value, for multiple districts


        :return:
        """

        G = fig2_graph()

        variables = list(G.vertices)
        districts = [frozenset(x) for x in G.districts]
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        q_vector = binary_nested.initialize_q_vector(intrinsic_dict)
        q = np.array(list(q_vector.values()))
        Ms, Ps = binary_nested.compute_all_M_and_P(
            G=G, intrinsic_dict=intrinsic_dict
        )
        district_bool_map = binary_nested.compute_district_bool_map(
            q_vector_keys=list(q_vector.keys()), districts=districts
        )
        Ms_map = dict(zip(districts, Ms))
        Ps_map = dict(zip(districts, Ps))

        for D_j in districts:
            M = Ms_map[D_j]
            P = Ps_map[D_j]

            for variable in sorted(D_j):
                # problem:
                # theta_bool_map maps variable to set of params
                # M and P map district to set of params
                keys = list(q_vector.keys())
                theta_bool_map = binary_nested.compute_theta_bool_map(
                    q_vector_keys=keys, variables=variables
                )
                theta_reindexed_bool_map = (
                    binary_nested.compute_theta_reindexed_bool_map(
                        q_vector_keys=keys, districts=districts
                    )
                )
                x_dis = q[district_bool_map[D_j]]
                A, b = binary_nested.construct_A_b(
                    variable=variable,
                    q=x_dis,
                    theta_reindexed_bool_map=theta_reindexed_bool_map,
                    M=M,
                    P=P,
                )

                x_var = q[theta_bool_map[variable]]
                result1 = binary_nested._compute_prob_single_district(
                    x_dis, M, P
                )
                result2 = A @ x_var - b
                np.testing.assert_array_almost_equal(result1, result2)

    def test_term_tail_length_on_verma(self):
        """
        Check that the combined tails in each term don't contain repeated variables.

        :return:
        """
        G = verma_graph(k=2)
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        districts = G.districts
        heads_tails_map = binary_nested.get_heads_tails_map(
            intrinsic_dict=intrinsic_dict
        )
        for district in districts:
            # partition_head_dict = binary_nested.compute_partition_head_dict(G=G, intrinsic_dict=intrinsic_dict, district=district)
            terms = binary_nested.compute_terms(
                district=district,
                intrinsic_dict=intrinsic_dict,
                heads_tails_map=heads_tails_map,
            )

            print(terms)
            for head, tail, value in terms:
                # Check that there are no repeats in the tail variables

                self.assertEqual(len(tail), len(set(tail)))

    def test_that_16_not_18_count_correct(self):
        """
        Check that the 16-not-18 graph has 16 parameters as predicted under the nested Markov model, not 18 parameters under the ordinary Markov model.
        :return:
        """
        G = sixteen_not_eighteen_graph()
        intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()
        q_vector = binary_nested.initialize_q_vector(intrinsic_dict)
        self.assertEqual(len(q_vector), 16)


class TestBinaryNestedModel(unittest.TestCase):
    def test_process_data_on_nonbinary_dataframe_fails(self):
        X = pd.DataFrame({"X1": [0, 0, 2, 1, 1], "X2": [0, 1, 0, 1, 2]})
        with pytest.raises(KeyError):
            result = binary_nested.process_data(X)

    def test_that_process_data_correctly_extracts_counts(self):
        X = pd.DataFrame({"X1": [0, 0, 1, 1, 1], "X2": [0, 1, 0, 1, 1]})
        result = binary_nested.process_data(X)
        expected = np.array([1, 1, 1, 2])
        np.testing.assert_array_equal(result, expected)

        X = pd.DataFrame({"X1": [0, 1, 1, 1, 1], "X2": [0, 1, 0, 1, 1]})
        result = binary_nested.process_data(X)
        expected = np.array([1, 0, 1, 3])
        np.testing.assert_array_equal(result, expected)

    def test_that_process_data_extracts_tabular_summary_correctly(self):
        X = pd.DataFrame(
            {"X2": [0, 0, 1, 1], "X1": [0, 1, 0, 1], "count": [1, 2, 3, 4]}
        )
        result = binary_nested.process_data(X, count_variable="count")
        expected = np.array([1, 3, 2, 4])
        np.testing.assert_array_equal(result, expected)



"""
Class for Undirected Graphs (UGs).
"""


from .cg import CG


class UG(CG):
    def __init__(self, vertices=[], ud_edges=set(), **kwargs):
        """
        Constructor.

        :param vertices: iterable of names of vertices.
        :param ud_edges: iterable of tuples of undirected edges i.e. (X, Y) = X - Y.
        """

        super().__init__(vertices=vertices, ud_edges=ud_edges, **kwargs)



import unittest

from ananke.graphs import CG


class TestCG(unittest.TestCase):
    def test_cg_boundary(self):
        vertices = ["X_1", "X_2", "X_3", "X_4"]
        di_edges = [("X_1", "X_2")]
        ud_edges = [("X_2", "X_3")]
        G = CG(vertices=vertices, di_edges=di_edges, ud_edges=ud_edges)

        boundary = G.boundary(["X_2"])

        self.assertEqual(boundary, {"X_1", "X_3"})



import functools
import logging
from collections import ChainMap

from pgmpy.factors.discrete import DiscreteFactor, TabularCPD

from ananke.inference.variable_elimination import variable_elimination

from ..graphs import ADMG, DAG

logger = logging.getLogger(__name__)


def compute_district_factor(graph: ADMG, obs_dist, fixing_order):
    """
    Compute the interventional distribution associated with a district (or equivalently, its fixing order)

    :param graph: Graph representing the problem
    :type graph: ananke.ADMG
    :param obs_dist: Probability distribution corresponding to the graph
    :type obs_dist: pgmpy.discrete.DiscreteFactor
    :param fixing_order: A fixing sequence for the implied district D
    """
    new_graph = graph.copy()
    curr_factor = obs_dist

    logger.info(
        f"implied district is {set(graph.vertices) - set(fixing_order)}"
    )

    for var in fixing_order:
        non_descendants = list(
            set(new_graph.vertices) - set(new_graph.descendants(var))
        )
        logger.info(f"fixing by q({var}|{', '.join(non_descendants)})")
        div_joint = curr_factor.marginalize(
            set(new_graph.descendants(var)) - {var}, inplace=False
        )
        if non_descendants:
            div_cond = div_joint.marginalize([var], inplace=False)
            div_factor = div_joint.divide(div_cond, inplace=False)
        else:
            div_factor = div_joint
        new_graph = new_graph.fix(var)

        curr_factor = curr_factor.divide(div_factor, inplace=False)

    return curr_factor


def get_obs_dist_from_net(net, graph):
    return variable_elimination(net, graph.vertices)


def estimate_effect_from_discrete_dist(
    oid, obs_dist, treatment_dict, outcome_dict
):
    """
    Performs the ID algorithm to identify a causal effect given a discrete
    probability distribution representing the observed data distribution.


    :param oid: Ananke OneLineID object
    :type oid: OneLineID
    :param obs_dist: observed data distribution
    :param treatment_dict: dictionary of treatment variables and values
    :param outcome_dict: dictionary of outcome variables and values
    """
    if not oid.id():
        return None

    factors = list()
    for district in sorted(oid.Gystar.districts):
        fixing_order = oid.fixing_orders[tuple(district)]
        factors.append(
            compute_district_factor(oid.graph, obs_dist, fixing_order)
        )

    intervened_distribution = functools.reduce(
        (lambda first, last: first.product(last, inplace=False)), factors
    )

    # construct the variable tuples here using itertools.product (Y* - Y) with A = a
    summed_vars = oid.ystar - set(outcome_dict)

    # compute the causal effect

    causal_effect = intervened_distribution.marginalize(
        summed_vars, inplace=False
    ).get_value(**dict(ChainMap(treatment_dict, outcome_dict)))

    return causal_effect



import unittest

from ananke.graphs import ADMG, IG


class TestIG(unittest.TestCase):
    def test_init(self):
        vertices = ["A", "B", "C"]
        di_edges = [("A", "B")]
        bi_edges = [("B", "C"), ("A", "C")]

        admg = ADMG(vertices=vertices, bi_edges=bi_edges, di_edges=di_edges)

        ig = IG(admg=admg)
        print(ig.bi_edges)

    def test_get_intrinsic_sets_three_var_graph(self):
        vertices = ["A1", "B", "C"]
        di_edges = [("A1", "B")]
        bi_edges = [("B", "C"), ("A1", "C")]

        admg = ADMG(vertices=vertices, bi_edges=bi_edges, di_edges=di_edges)

        ig = IG(admg=admg)
        intrinsic_sets = ig.get_intrinsic_sets()
        print(intrinsic_sets)
        self.assertEqual(
            {
                frozenset({"C"}),
                frozenset({"C", "B", "A1"}),
                frozenset({"C", "A1"}),
                frozenset({"A1"}),
                frozenset({"B"}),
            },
            intrinsic_sets,
        )

    def test_get_intrinsic_sets_complete_bidirected_three_var_graph(self):
        vertices = ["A", "B", "C"]
        di_edges = []
        bi_edges = [("A", "B"), ("B", "C"), ("A", "C")]
        admg = ADMG(vertices=vertices, bi_edges=bi_edges, di_edges=di_edges)
        ig = IG(admg=admg)
        intrinsic_sets = ig.get_intrinsic_sets()
        print(intrinsic_sets)
        self.assertEqual(
            {
                frozenset({"A"}),
                frozenset({"B"}),
                frozenset({"C"}),
                frozenset({"A", "B"}),
                frozenset({"A", "C"}),
                frozenset({"B", "C"}),
                frozenset({"A", "B", "C"}),
            },
            intrinsic_sets,
        )

    def test_get_intrinsic_sets_four_var_graph(self):
        vertices = ["A", "B", "C", "D"]
        di_edges = [("A", "B"), ("B", "C"), ("C", "D")]
        bi_edges = [("A", "C"), ("A", "D")]

        admg = ADMG(vertices=vertices, bi_edges=bi_edges, di_edges=di_edges)
        ig = IG(admg=admg)
        intrinsic_sets = ig.get_intrinsic_sets()

        self.assertEqual(
            {
                frozenset({"A"}),
                frozenset({"B"}),
                frozenset({"C"}),
                frozenset({"D"}),
                frozenset({"A", "C"}),
                frozenset({"A", "C", "D"}),
            },
            intrinsic_sets,
        )

    def test_get_intrinsic_sets_five_var(self):
        vertices = ["A", "B", "C", "D", "Y"]
        di_edges = [("A", "B"), ("A", "C"), ("B", "C"), ("C", "D"), ("D", "Y")]
        bi_edges = [("A", "C"), ("B", "D"), ("C", "Y")]
        admg = ADMG(vertices=vertices, bi_edges=bi_edges, di_edges=di_edges)
        ig = IG(admg=admg)
        intrinsic_sets = ig.get_intrinsic_sets()
        truth = {
            frozenset({"D"}),
            frozenset({"C", "A"}),
            frozenset({"Y"}),
            frozenset({"D", "B"}),
            frozenset({"B"}),
            frozenset({"A"}),
            frozenset({"C", "Y", "A"}),
        }
        self.assertEqual(truth, intrinsic_sets)
        print(intrinsic_sets)
        print("-" * 20)
        print(ig.get_heads_tails())

    def test_check_fixing_order(self):
        vertices = ["A", "B", "C"]
        di_edges = [("A", "B")]
        bi_edges = [("B", "C"), ("A", "C")]

        admg = ADMG(vertices=vertices, bi_edges=bi_edges, di_edges=di_edges)

        ig = IG(admg=admg)
        intrinsic_sets = ig.get_intrinsic_sets()
        print(intrinsic_sets)
        fixing_order = ig.iset_fixing_order_map
        print(fixing_order)
        self.assertEqual(["C", "A"], fixing_order[frozenset(["B"])])

    def test_get_intrinsic_sets_in_cadmg(self):
        vertices = ["A", "B", "C"]
        di_edges = [("A", "B")]
        bi_edges = [("B", "C"), ("A", "C")]

        admg = ADMG(vertices=vertices, bi_edges=bi_edges, di_edges=di_edges)
        admg.fix(["B"])
        ig = IG(admg=admg)
        intrinsic_sets = ig.get_intrinsic_sets()
        self.assertEqual(
            {frozenset(["A"]), frozenset(["C"]), frozenset(["A", "C"])},
            intrinsic_sets,
        )


if __name__ == "__main__":
    unittest.main()



import sympy as sp

from ananke.factors import SymCPD
from ananke.graphs import DAG
from ananke.identification import oracle
from ananke.models.bayesian_network import BayesianNetwork


def test_compute_effect_from_discrete_model():

    u1 = sp.Symbol("u1")
    a1 = sp.Symbol("a1")
    a2 = sp.Symbol("a2")
    y1 = sp.Symbol("y1")
    y2 = sp.Symbol("y2")
    y3 = sp.Symbol("y3")
    y4 = sp.Symbol("y4")
    cpd_u = SymCPD("U", 2, sp.Array([[u1], [1 - u1]]))
    cpd_a = SymCPD("A", 2, sp.Array([[a1, a2], [1 - a1, 1 - a2]]), ["U"], [2])

    cpd_y = SymCPD(
        "Y",
        2,
        sp.Array([[y1, y2, y3, y4], [1 - y1, 1 - y2, 1 - y3, 1 - y4]]),
        ["A", "U"],
        [2, 2],
    )
    graph = DAG(
        vertices={"A": 2, "U": 2, "Y": 2},
        di_edges=[("A", "Y"), ("U", "A"), ("U", "Y")],
    )
    cpds = {"A": cpd_a, "Y": cpd_y, "U": cpd_u}
    net = BayesianNetwork(graph, cpds)

    result = oracle.compute_effect_from_discrete_model(net, {"A": 0}, {"Y": 0})
    truth = y1 * u1 + y2 * (1 - u1)
    assert sp.simplify(result - truth) == 0



import unittest

from ananke.utils import powerset


class TestUtils(unittest.TestCase):
    def test_powerset(self):

        self.assertEqual(
            list(powerset([1, 2, 3])),
            [(), (1,), (2,), (3,), (1, 2), (1, 3), (2, 3), (1, 2, 3)],
        )


if __name__ == "__main__":
    unittest.main()



import copy
import itertools
import unittest

import numpy as np
import pytest
import sympy as sp
from numpy.testing import assert_array_equal
from pgmpy.factors.discrete import TabularCPD
from pgmpy.inference import VariableElimination

from ananke import identification
from ananke.graphs import ADMG, DAG, Vertex
from ananke.inference.variable_elimination import variable_elimination
from ananke.models import BayesianNetwork, create_symbolic_cpds


class TestOneLine(unittest.TestCase):
    def test_id_graph(self):
        """
        Test that ADMG Y(a) is identified
        """

        vertices = ["A", "B", "C", "D", "Y"]
        di_edges = [
            ("A", "B"),
            ("A", "D"),
            ("B", "C"),
            ("C", "Y"),
            ("B", "D"),
            ("D", "Y"),
        ]
        bi_edges = [("A", "C"), ("B", "Y"), ("B", "D")]
        G = ADMG(vertices, di_edges, bi_edges)
        one_id = identification.OneLineID(G, ["A"], ["Y"])
        one_id.draw_swig()
        self.assertTrue(one_id.id())
        self.assertEqual({"Y", "C", "D", "B"}, one_id.ystar)
        one_id.export_intermediates()

    def test_BMAY_graph(self):
        vertices = ["B", "M", "A", "Y"]
        di_edges = [("B", "M"), ("M", "A"), ("A", "Y")]
        bi_edges = [("B", "A"), ("B", "Y")]
        G = ADMG(vertices, di_edges, bi_edges)
        one_id = identification.OneLineID(G, ["A"], ["Y"])
        one_id.draw_swig()
        self.assertTrue(one_id.id())

    def test_two_var_id_graph(self):
        vertices = ["A", "D", "C", "Y"]
        di_edges = [("D", "Y"), ("D", "A"), ("A", "Y"), ("C", "A"), ("C", "Y")]
        bi_edges = []
        G = ADMG(vertices, di_edges, bi_edges)
        one_id = identification.OneLineID(G, ["A"], ["Y", "D"])
        self.assertEqual({"C", "D", "Y"}, one_id.ystar)
        self.assertTrue(one_id.id())

    def test_non_id_graph(self):
        """
        Test that Y(a,b) is not identified
        """

        # non ID test
        vertices = ["A", "B", "C", "D", "Y"]
        di_edges = [
            ("A", "B"),
            ("A", "D"),
            ("B", "C"),
            ("C", "Y"),
            ("B", "D"),
            ("D", "Y"),
        ]
        bi_edges = [("A", "C"), ("B", "Y"), ("B", "D")]
        G = ADMG(vertices, di_edges, bi_edges)
        one_id = identification.OneLineID(G, ["A", "B"], ["Y"])
        self.assertFalse(one_id.id())

    def test_functional(self):
        vertices = ["M", "A", "Y"]
        di_edges = [("A", "M"), ("M", "Y")]
        bi_edges = [("A", "Y")]
        G = ADMG(vertices, di_edges, bi_edges)
        one_id = identification.OneLineID(G, ["A"], ["Y"])
        functional = one_id.functional()
        self.assertEqual("ΣM ΦAY(p(V);G) ΦAM(p(V);G) ", functional)

    def test_id_after_fixing_fails_properly(self):
        vertices = ["X1", "X2", "W", "Y"]
        di_edges = [("X1", "W"), ("W", "Y"), ("X2", "Y")]
        bi_edges = [("X1", "X2"), ("X1", "W"), ("X2", "Y")]
        G = ADMG(vertices, di_edges=di_edges, bi_edges=bi_edges)
        G.fix(["X1"])

        # This should not be identified as X_2 -> Y <-> X_2 is a bow-arc graph
        self.assertFalse(
            identification.OneLineID(G, treatments=["X2"], outcomes=["Y"]).id()
        )

        # Check that NotIdentifiedError is raised if we try to obtain functional
        with self.assertRaises(identification.NotIdentifiedError):
            identification.OneLineID(
                G, treatments=["X2"], outcomes=["Y"]
            ).functional()

    def test_id_fails_correctly(self):

        vertices = ["X1", "X2", "W", "Y"]
        di_edges = [("X1", "W"), ("W", "Y"), ("X2", "Y")]
        bi_edges = [("X1", "X2"), ("X1", "W"), ("X2", "Y")]
        G = ADMG(vertices, di_edges=di_edges, bi_edges=bi_edges)
        G.fix(["X1"])

        # This should not be identified as X_2 -> Y <-> X_2 is a bow-arc graph
        self.assertFalse(
            identification.OneLineID(G, treatments=["X2"], outcomes=["Y"]).id()
        )

        # Check that NotIdentifiedError is raised if we try to obtain functional
        with self.assertRaises(identification.NotIdentifiedError):
            identification.OneLineID(
                G, treatments=["X2"], outcomes=["Y"]
            ).functional()


class TestOneLineGID(unittest.TestCase):
    def test_is_id(self):
        vertices = ["X_1", "X_2", "W", "Y"]
        di_edges = [("X_1", "W"), ("W", "Y"), ("X_2", "Y")]
        bi_edges = [("X_1", "W"), ("X_2", "Y"), ("X_1", "X_2")]
        G = ADMG(vertices, di_edges, bi_edges)
        interventions = ["X_1", "X_2"]
        outcomes = ["Y"]
        ol = identification.OneLineGID(G, interventions, outcomes)
        status = ol.id()

        self.assertFalse(status)

        experiments = [G.copy().fix("X_1"), G.copy().fix("X_2")]
        second = ol.id(experiments)
        self.assertTrue(second)

    def test_functional(self):
        vertices = ["X_1", "X_2", "W", "Y"]
        di_edges = [("X_1", "W"), ("W", "Y"), ("X_2", "Y")]
        bi_edges = [("X_1", "W"), ("X_2", "Y"), ("X_1", "X_2")]
        G = ADMG(vertices, di_edges, bi_edges)
        interventions = ["X_1", "X_2"]
        outcomes = ["Y"]
        ol = identification.OneLineGID(G, interventions, outcomes)
        functional = ol.functional([G.copy().fix("X_1"), G.copy().fix("X_2")])
        self.assertEqual(
            "ΣW ΦX_2,Y p(W,X_2,Y | do(X_1))ΦX_1,W p(W,X_1,Y | do(X_2))",
            functional,
        )

    def test_is_id_chain(self):
        vertices = ["A", "X", "W", "Y"]
        di_edges = [("A", "X"), ("X", "W"), ("W", "Y")]
        bi_edges = [("X", "W"), ("W", "Y")]
        G = ADMG(vertices, di_edges, bi_edges)
        interventions = ["A"]
        outcomes = ["Y"]
        ol = identification.OneLineGID(G, interventions, outcomes)
        status = ol.id([G.copy().fix(["A", "X"]), G.copy().fix({"A", "Y"})])
        self.assertFalse(status)

        vertices = ["A", "X", "Y"]
        di_edges = [("A", "X"), ("X", "Y")]
        bi_edges = [("X", "Y")]
        G = ADMG(vertices, di_edges, bi_edges)
        interventions = ["A"]
        outcomes = ["Y"]
        ol = identification.OneLineGID(G, interventions, outcomes)
        status = ol.id([G.copy().fix({"A", "X"}), G.copy().fix({"A", "Y"})])
        self.assertFalse(status)

        vertices = ["A", "W", "Y"]
        di_edges = [("A", "W"), ("W", "Y")]
        bi_edges = [("A", "W"), ("W", "Y")]
        G = ADMG(vertices, di_edges, bi_edges)
        interventions = ["A"]
        outcomes = ["Y"]
        ol = identification.OneLineGID(G, interventions, outcomes)
        status = ol.id([G.copy().fix({"A"}), G.copy().fix({"A", "Y"})])
        self.assertTrue(status)

    def test_that_gid_is_correct(self):
        vertices = ["X1", "X2", "W", "Y"]
        di_edges = [("X1", "W"), ("W", "Y"), ("X2", "Y")]
        bi_edges = [("X1", "X2"), ("X1", "W"), ("X2", "Y")]
        G = ADMG(vertices, di_edges, bi_edges)

        G1 = ADMG(vertices, di_edges, bi_edges)
        G1.fix(["X1"])

        G2 = ADMG(vertices, di_edges, bi_edges)
        G2.fix(["X2"])
        result = identification.OneLineGID(
            graph=G, treatments=["X1", "X2"], outcomes=["Y"]
        ).id(experiments=[G1, G2])


class TestOnelineAID(unittest.TestCase):
    def test_oneline_aid(self):
        vertices = ["X1", "X2", "W", "Y"]
        di_edges = [("X1", "W"), ("W", "Y"), ("X2", "Y")]
        bi_edges = [("X1", "X2"), ("X1", "W"), ("X2", "Y")]
        G = ADMG(vertices, di_edges, bi_edges)

        vertices = ["X1", "W"]
        di_edges = [("X1", "W")]
        bi_edges = [("X1", "W")]
        G1 = ADMG(vertices, di_edges, bi_edges)
        G1.fix(["X1"])

        vertices = ["X1", "X2", "W", "Y"]
        di_edges = [("X1", "W"), ("W", "Y"), ("X2", "Y")]
        bi_edges = [("X1", "X2"), ("X1", "W"), ("X2", "Y")]
        G2 = ADMG(vertices, di_edges, bi_edges)
        G2.fix(["X2"])

        interventions = ["X1", "X2"]
        outcomes = ["Y"]
        ol = identification.OneLineAID(G, interventions, outcomes)

        experiments = [G1, G2]

        self.assertTrue(ol.id(experiments=experiments))
        self.assertEqual(
            "ΣW  p(W | do(X1))ΦX1,W p(W,X1,Y | do(X2))",
            ol.functional(experiments),
        )


class TestValidWitness:
    def test_ananke_witness(self, bow_arc_bayesian_network_models):
        net_1, net_2 = bow_arc_bayesian_network_models
        identification.one_line.assert_valid_witness(
            net_1, net_2, ["Y", "A"], {"A": 0}, ["Y"]
        )

    def test_pgmpy_witness(self, bow_arc_bayesian_network_models):
        net_1, net_2 = bow_arc_bayesian_network_models
        net_1 = net_1.to_pgmpy()
        net_2 = net_2.to_pgmpy()
        identification.one_line.assert_valid_witness(
            net_1, net_2, ["Y", "A"], {"A": 0}, ["Y"]
        )
